import ctypes
from ctypes import wintypes
from typing import Optional

user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32
gdi32 = ctypes.windll.gdi32

from phardwareitk.GUI.PheonixIon.types import *
from phardwareitk.GUI.pheonix_ion import GPU_API, PIonContext
from phardwareitk.Extensions.Windows import *
from phardwareitk.GPU._base import BaseGPUD

user32.DefWindowProcW.argtypes = [HWND, UINT, WPARAM, LPARAM]
user32.DefWindowProcW.restype  = LRESULT
user32.PeekMessageW.argtypes = [ctypes.POINTER(MSG), HWND, UINT, UINT, UINT]
user32.PeekMessageW.restype  = BOOL
user32.TranslateMessage.argtypes = [ctypes.POINTER(MSG)]
user32.TranslateMessage.restype  = BOOL
user32.DispatchMessageW.argtypes = [ctypes.POINTER(MSG)]
user32.DispatchMessageW.restype  = LRESULT

gdi32.GetStockObject.restype = HGDIOBJ
gdi32.GetStockObject.argtypes = [ctypes.c_int]

_windows_events = []

class PIWin32Flags:
    def __init__(
        self,
        exStyle: int = 0,
        className="PheonixIonWindow",
        windowName="Pheonix Ion",
        style: int = WS_OVERLAPPEDWINDOW,
        x: int = 100,
        y: int = 100,
        width: int = 800,
        height: int = 600,
        parent: Optional[HWND] = None,
        menu: Optional[HMENU] = None,
        instance: Optional[HINSTANCE] = None,
        param: Optional[LPVOID] = None,
    ) -> None:
        self.exStyle = exStyle
        self.className = className
        self.windowName = windowName
        self.style = style
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.parent = parent
        self.menu = menu
        self.instance = instance if instance else kernel32.GetModuleHandleW(None)
        self.param = param

    def to_c(self):
        return (
            self.exStyle,
            self.className,
            self.windowName,
            self.style,
            self.x,
            self.y,
            self.width,
            self.height,
            self.parent,
            self.menu,
            self.instance,
            self.param,
        )


def _wnd_proc(hwnd, msg, wparam, lparam):
    global _windows_events
    # --- Window lifecycle ---
    if msg == WM_CREATE:
        _windows_events.append(PIonEvent("CREATE"))
        return 0
    elif msg == WM_DESTROY:
        _windows_events.append(PIonEvent("QUIT"))
        user32.PostQuitMessage(0)
        return 0
    elif msg == WM_CLOSE:
        _windows_events.append(PIonEvent("CLOSE", destroyed=False))
        return 0
    elif msg == WM_SHOWWINDOW:
        if wparam:  # TRUE means shown
            _windows_events.append(PIonEvent("SHOW"))
        else:
            _windows_events.append(PIonEvent("HIDE"))
        return 0
    elif msg == WM_SIZE:
        width = lparam & 0xFFFF
        height = (lparam >> 16) & 0xFFFF
        size_type = wparam
        if size_type == 1:  # minimized
            _windows_events.append(PIonEvent("MINIMIZE"))
        elif size_type == 2:  # maximized
            _windows_events.append(PIonEvent("MAXIMIZE"))
        elif size_type == 0:  # restored
            _windows_events.append(PIonEvent("RESTORE"))
        _windows_events.append(PIonEvent("RESIZE", width=width, height=height))
        return 0
    elif msg == WM_MOVE:
        x = lparam & 0xFFFF
        y = (lparam >> 16) & 0xFFFF
        _windows_events.append(PIonEvent("MOVE", x=x, y=y))
        return 0

    # --- Mouse ---
    elif msg == WM_MOUSEMOVE:
        x = lparam & 0xFFFF
        y = (lparam >> 16) & 0xFFFF
        _windows_events.append(PIonEvent("MOUSEMOVE", x=x, y=y))
        return 0
    elif msg == WM_LBUTTONDOWN:
        _windows_events.append(PIonEvent("LEFT_DOWN"))
        return 0
    elif msg == WM_LBUTTONUP:
        _windows_events.append(PIonEvent("LEFT_UP"))
        return 0
    elif msg == WM_RBUTTONDOWN:
        _windows_events.append(PIonEvent("RIGHT_DOWN"))
        return 0
    elif msg == WM_RBUTTONUP:
        _windows_events.append(PIonEvent("RIGHT_UP"))
        return 0
    elif msg == WM_MBUTTONDOWN:
        _windows_events.append(PIonEvent("MIDDLE_DOWN"))
        return 0
    elif msg == WM_MBUTTONUP:
        _windows_events.append(PIonEvent("MIDDLE_UP"))
        return 0
    elif msg == WM_MOUSEWHEEL:
        delta = ctypes.c_short((wparam >> 16) & 0xFFFF).value
        _windows_events.append(PIonEvent("SCROLL", delta=delta))
        return 0
    elif msg == WM_MOUSELEAVE:
        _windows_events.append(PIonEvent("MOUSELEAVE"))
        return 0
    elif msg == WM_MOUSEHOVER:
        _windows_events.append(PIonEvent("MOUSEENTER"))
        return 0

    # --- Keyboard ---
    elif msg == WM_KEYDOWN:
        _windows_events.append(PIonEvent("KEYDOWN", keycode=wparam))
        return 0
    elif msg == WM_KEYUP:
        _windows_events.append(PIonEvent("KEYUP", keycode=wparam))
        return 0
    elif msg == WM_CHAR:
        _windows_events.append(PIonEvent("KEYPRESS", char=chr(wparam)))
        return 0
    elif msg == WM_IME_COMPOSITION:
        _windows_events.append(PIonEvent("IME_STARTCOMPOSITION"))
        return 0
    elif msg == WM_IME_ENDCOMPOSITION:
        _windows_events.append(PIonEvent("IME_ENDCOMPOSITION"))
        return 0

    # --- Other / system ---
    elif msg == WM_DISPLAYCHANGE:
        _windows_events.append(PIonEvent("DISPLAYCHANGE"))
        return 0
    elif msg == WM_POWERBROADCAST:
        _windows_events.append(PIonEvent("POWERBROADCAST"))
        return 0
    elif msg == WM_TIMER:
        _windows_events.append(PIonEvent("TIMER", timer_id=wparam))
        return 0

    return user32.DefWindowProcW(hwnd, msg, wparam, lparam)

_wnd_proc_callback = WNDPROC(_wnd_proc)

def create_window(
    title: str = "Pheonix Ion",
    width: int = 800,
    height: int = 600,
    flags: PIWin32Flags = None,
) -> HWND:
    """Create a Win32 window and return its handle."""
    hInstance = kernel32.GetModuleHandleW(None)
    className = "PheonixIonWindow"

    # Register class
    wndclass = WNDCLASS()
    wndclass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC
    wndclass.lpfnWndProc = _wnd_proc_callback
    wndclass.cbClsExtra = 0
    wndclass.cbWndExtra = 0
    wndclass.hInstance = hInstance
    wndclass.hIcon = user32.LoadIconW(None, IDI_APPLICATION)
    wndclass.hCursor = user32.LoadCursorW(None, IDC_ARROW)
    wndclass.hbrBackground = gdi32.GetStockObject(WHITE_BRUSH)
    wndclass.lpszMenuName = None
    wndclass.lpszClassName = className

    if not flags:
        flags = PIWin32Flags(
            0,
            className,
            title,
            WS_OVERLAPPEDWINDOW,
            100,
            100,
            width,
            height,
            None,
            None,
            hInstance,
            None,
        )

    if not user32.RegisterClassW(ctypes.byref(wndclass)):
        raise ctypes.WinError()

    hwnd = user32.CreateWindowExW(*flags.to_c())
    return hwnd


def show_window(hwnd: HWND) -> None:
    user32.ShowWindow(hwnd, 1)
    user32.UpdateWindow(hwnd)

def hide_window(hwnd: HWND) -> None:
    user32.ShowWindow(hwnd, 0)
    user32.UpdateWindow(hwnd)

def set_window_title(hwnd: HWND, title: str) -> None:
    user32.SetWindowTextW(hwnd, title)

def destroy_window(hwnd: HWND) -> None:
    user32.DestroyWindow(hwnd)

def poll_events(hwnd: HWND):
    msg = MSG()
    ctypes.memset(ctypes.byref(msg), 0, ctypes.sizeof(msg))
    events = _windows_events.copy()
    _windows_events.clear()
    while user32.PeekMessageW(ctypes.byref(msg), hwnd, 0, 0, 1):
        user32.TranslateMessage(ctypes.byref(msg))
        user32.DispatchMessageW(ctypes.byref(msg))
    return events


def is_window_alive(hwnd: HWND) -> bool:
    return user32.IsWindow(hwnd) != 0

def get_gpu(hwnd: HWND, api: Optional[str], driver: Optional[BaseGPUD]) -> GPU_API:
    if not hwnd:
        raise RuntimeError("Cannot create context for a invalid window!")

    gpu = GPU_API(api, driver)
    gpu.driver.init(None, hwnd, create_and_attach_ctx=False)
    return gpu

def attach_gpu(hwnd: HWND, gpu: GPU_API) -> PIonContext:
    return gpu.driver.create_context(None, hwnd)


"""This file includes everything for windows platform"""

import sys
from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions.C import *

from typing import * # type: ignore
import ctypes

# Winuser.h -> Class Styles
CS_BYTEALIGNCLIENT = 0x1000
CS_BYTEALIGNWINDOW = 0x2000
CS_CLASSDC = 0x0040
CS_DBLCLKS = 0x0008
CS_DROPSHADOW = 0x00020000
CS_GLOBALCLASS = 0x4000
CS_HREDRAW = 0x0002
CS_NOCLOSE = 0x0200
CS_OWNDC = 0x0020
CS_PARENTDC = 0x0080
CS_SAVEBITS = 0x0800
CS_VREDRAW = 0x0001

# winuser.h -> Icons
IDI_APPLICATION = 32512
IDI_ERROR = 32513
IDI_QUESTION = 32514
IDI_WARNING = 32515
IDI_INFORMATION = 32516
IDI_WINLOGO = 32517
IDI_SHIELD = 32518

# winuser.h -> Cursors
IDC_ARROW = 32512
IDC_IBEAM = 32513
IDC_WAIT = 32514
IDC_CROSS = 32515
IDC_UPARROW = 32516
IDC_SIZENWSE = 32642
IDC_SIZENESW = 32643
IDC_SIZEWE = 32644
IDC_SIZENS = 32645
IDC_SIZEALL = 32646
IDC_NO = 32648
IDC_HAND = 32649
IDC_APPSTARTING = 32650
IDC_HELP = 32651
IDC_PIN = 32671
IDC_PERSON = 32672

# wingdi -> GetStockObject
# These are copied from win32con (only these ones not others)
WHITE_BRUSH = 0
LTGRAY_BRUSH = 1
GRAY_BRUSH = 2
DKGRAY_BRUSH = 3
BLACK_BRUSH = 4
NULL_BRUSH = 5
HOLLOW_BRUSH = NULL_BRUSH
WHITE_PEN = 6
BLACK_PEN = 7
NULL_PEN = 8
OEM_FIXED_FONT = 10
ANSI_FIXED_FONT = 11
ANSI_VAR_FONT = 12
SYSTEM_FONT = 13
DEVICE_DEFAULT_FONT = 14
DEFAULT_PALETTE = 15

# winuser.h -> Size and location
CW_USEDEFAULT = -2147483648

# winuser.h -> Show Window
SW_HIDE = 0
SW_SHOWNORMAL = 1
SW_NORMAL = 1
SW_SHOWMINIMIZED = 2
SW_SHOWMAXIMIZED = 3
SW_MAXIMIZED = 3
SW_SHOWNOACTIVATE = 4
SW_SHOW = 5
SW_MINIMIZE = 6
SW_SHOWMINNOACTIVE = 7
SW_SHOWNA = 8
SW_RESTORE = 9
SW_SHOWDEFAULT = 10
SW_FORCEMINIMIZE = 11

# wingdi -> Painting and Drawing Messages
WM_DISPLAYCHANGE = 126
WM_ERASEBKGND = 20
WM_NCPAINT = 133
WM_PAINT = 15
WM_PRINT = 791
WM_PRINTCLIENT = 792
WM_SETREDRAW = 11
WM_SYNCPAINT = 136
# Windows Messages
WM_CREATE = 0x0001
WM_DESTROY = 0x0002
WM_CLOSE = 0x0010
WM_SHOWWINDOW = 0x0018
WM_SIZE = 0x0005
WM_MOVE = 0x0003
WM_MOUSEMOVE = 0x0200
WM_LBUTTONDOWN = 0x0201
WM_LBUTTONUP = 0x0202
WM_RBUTTONDOWN = 0x0204
WM_RBUTTONUP = 0x0205
WM_MBUTTONDOWN = 0x0207
WM_MBUTTONUP = 0x0208
WM_MOUSEWHEEL = 0x020A
WM_MOUSELEAVE = 0x02A3
WM_MOUSEHOVER = 0x02A1
WM_KEYDOWN = 0x0100
WM_KEYUP = 0x0101
WM_CHAR = 0x0102
WM_IME_COMPOSITION = 0x010F
WM_IME_ENDCOMPOSITION = 0x010E
WM_DISPLAYCHANGE = 0x007E
WM_POWERBROADCAST = 0x0218
WM_TIMER = 0x0113

# winuser.h -> PeekMessage Constants Part-1
PM_NOREMOVE = 0x0000
PM_REMOVE = 0x0001
PM_NOYIELD = 0x0002

# Windows styles
WS_BORDER = 0x00800000
WS_CAPTION = 0x00C00000
WS_CHILD = 0x40000000
WS_CHILDWINDOW = WS_CHILD
WS_CLIPCHILDREN = 0x02000000
WS_CLIPSIBLINGS = 0x04000000
WS_DISABLED = 0x08000000
WS_DLGFRAME = 0x00400000
WS_GROUP = 0x00020000
WS_HSCROLL = 0x00100000
WS_ICONIC = 0x20000000
WS_MAXIMIZE = 0x01000000
WS_MAXIMIZEBOX = 0x00010000
WS_MINIMIZE = 0x20000000
WS_MINIMIZEBOX = 0x00020000
WS_OVERLAPPED = 0x00000000
WS_SYSMENU = 0x00080000
WS_THICKFRAME = 0x00040000
WS_OVERLAPPEDWINDOW = (WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)
WS_POPUP = 0x80000000
WS_POPUPWINDOW = (WS_POPUP | WS_BORDER | WS_SYSMENU)
WS_SIZEBOX = WS_THICKFRAME
WS_TABSTOP = 0x00010000
WS_TILED = 0x00000000
WS_TILEDWINDOW = (WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)
WS_VISIBLE = 0x10000000
WS_VSCROLL = 0x00200000

# Handle Types
HANDLE = ctypes.c_void_p

HACCEL = HANDLE
HBITMAP = HANDLE
HBRUSH = HANDLE
HCOLORSPACE = HANDLE
HDC = HANDLE
HDESK = HANDLE
HDWP = HANDLE
HENHMETAFILE = HANDLE
HFONT = HANDLE
HGDIOBJ = HANDLE
HGLOBAL = HANDLE
HHOOK = HANDLE
HICON = HANDLE
HCURSOR = HANDLE
HINSTANCE = HANDLE
HKEY = HANDLE
HKL = HANDLE
HLOCAL = HANDLE
HMENU = HANDLE
HMETAFILE = HANDLE
HMODULE = HANDLE
HMONITOR = HANDLE
HPALETTE = HANDLE
HPEN = HANDLE
HRGN = HANDLE
HRSRC = HANDLE
HSTR = HANDLE
HTASK = HANDLE
HWINSTA = HANDLE
HWND = HANDLE
SC_HANDLE = HANDLE
SERVICE_STATUS_HANDLE = HANDLE

# Types
LPCOLESTR = LPOLESTR = OLESTR = ctypes.c_wchar_p
LPCWSTR = LPWSTR = ctypes.c_wchar_p
LPCSTR = LPSTR = ctypes.c_char_p
LPCVOID = LPVOID = ctypes.c_void_p
BYTE = ctypes.c_ubyte
WORD = ctypes.c_ushort
DWORD = ctypes.c_ulong
CHAR = ctypes.c_char
WCHAR = ctypes.c_wchar
UINT = ctypes.c_uint
INT = ctypes.c_int
DOUBLE = ctypes.c_double
FLOAT = ctypes.c_float
BOOLEAN = BYTE
BOOL = ctypes.c_int
ULONG = ctypes.c_ulong
LONG = ctypes.c_long
USHORT = ctypes.c_ushort
SHORT = ctypes.c_short

# Pointer Types
WPARAM  = ctypes.c_uint32
LPARAM  = ctypes.c_int32
LRESULT = ctypes.c_int32

if ctypes.sizeof(ctypes.c_void_p) == 8:
    WPARAM = ctypes.c_uint64
    LPARAM = ctypes.c_int64
    LRESULT = ctypes.c_int64

# Special
WNDPROC = ctypes.WINFUNCTYPE(LRESULT, HWND, UINT, WPARAM, LPARAM)

# Structs
class POINT(ctypes.Structure):
    _fields_ = [("x", LONG), ("y", LONG)]

class MSG(ctypes.Structure):
    _fields_ = [("hWnd", HWND), ("message", UINT), ("wParam", WPARAM), ("lParam", LPARAM), ("time", DWORD), ("pt", POINT)]

    #if ctypes.sizeof(ctypes.c_void_p) == 8: # 64-bit
    #    _fields_.append(("lPrivate", DWORD))

class WNDCLASS(ctypes.Structure):
    _fields_ = [
        ("style", UINT),
        ("lpfnWndProc", WNDPROC),
        ("cbClsExtra", ctypes.c_int),
        ("cbWndExtra", ctypes.c_int),
        ("hInstance", HINSTANCE),
        ("hIcon", HICON),
        ("hCursor", HCURSOR),
        ("hbrBackground", HBRUSH),
        ("lpszMenuName", LPCWSTR),
        ("lpszClassName", LPCWSTR),
    ]


# X11 YOU MADE MY LIFE A LIVING NIGHTMARE!!!!!!
# TOTAL HOURS WASTED: 9

import ctypes
import os
from ctypes import c_int, c_ulong, c_char_p, c_void_p, POINTER, c_long, c_uint, cdll

from phardwareitk.GPU._base import BaseGPUD
from phardwareitk.GUI.PheonixIon.types import *
from phardwareitk.GUI.pheonix_ion import GPU_API, PIonContext

libX11 = ctypes.cdll.LoadLibrary("libX11.so.6")

# Define opaque X11 types
Display = c_void_p
Window = c_ulong

# Core X types
# Not organised looks better as a dev, truely saying!
Atom = ctypes.c_ulong
Window = ctypes.c_ulong
Drawable = ctypes.c_ulong
Time = ctypes.c_ulong
XPointer = ctypes.c_char_p
Bool = ctypes.c_int
Colormap = ctypes.c_ulong
VisualID = ctypes.c_ulong

libX11.XCreateSimpleWindow.argtypes = [
    ctypes.c_void_p,  # Display*
    ctypes.c_ulong,   # Window (parent)
    ctypes.c_int,     # x
    ctypes.c_int,     # y
    ctypes.c_uint,    # width
    ctypes.c_uint,    # height
    ctypes.c_uint,    # border width
    ctypes.c_ulong,   # border color
    ctypes.c_ulong    # background color
]

class XSetWindowAttributes(ctypes.Structure):
    _fields_ = [
        ("background_pixmap", c_ulong),
        ("background_pixel", c_ulong),
        ("border_pixmap", c_ulong),
        ("border_pixel", c_ulong),
        ("bit_gravity", c_int),
        ("win_gravity", c_int),
        ("backing_store", c_int),
        ("backing_planes", c_ulong),
        ("backing_pixel", c_ulong),
        ("save_under", Bool),
        ("event_mask", c_long),
        ("do_not_propagate_mask", c_long),
        ("override_redirect", Bool),
        ("colormap", Colormap),
        ("cursor", c_ulong),
    ]

class XWindowAttributes(ctypes.Structure):
    _fields_ = [
        ("x", c_int),
        ("y", c_int),
        ("width", c_int),
        ("height", c_int),
        ("border_width", c_int),
        ("depth", c_int),
        ("visual", c_void_p),
        ("root", Window),
        ("class_", c_int),
        ("bit_gravity", c_int),
        ("win_gravity", c_int),
        ("backing_store", c_int),
        ("backing_planes", c_ulong),
        ("backing_pixel", c_ulong),
        ("save_under", Bool),
        ("colormap", Colormap),
        ("map_installed", Bool),
        ("map_state", c_int),
        ("all_event_masks", c_long),
        ("your_event_mask", c_long),
        ("do_not_propagate_mask", c_long),
        ("override_redirect", Bool),
        ("screen", c_void_p),
    ]

class XVisualInfo(ctypes.Structure):
    _fields_ = [
        ("visual", ctypes.c_void_p),
        ("visualid", ctypes.c_ulong),
        ("screen", ctypes.c_int),
        ("depth", ctypes.c_int),
        ("class_", ctypes.c_int),
        ("red_mask", ctypes.c_ulong),
        ("green_mask", ctypes.c_ulong),
        ("blue_mask", ctypes.c_ulong),
        ("colormap_size", ctypes.c_int),
        ("bits_per_rgb", ctypes.c_int),
    ]

libX11.XCreateSimpleWindow.restype = ctypes.c_ulong
libX11.XCreateWindow.argtypes = [
    Display, Window,
    c_int, c_int, c_uint, c_uint, c_uint,
    c_int, c_uint, c_void_p,
    c_ulong, POINTER(XSetWindowAttributes)
]
libX11.XCreateWindow.restype = c_ulong
libX11.XDefaultScreen.argtypes = [Display]
libX11.XDefaultScreen.restype = c_int

libX11.XDefaultVisual.argtypes = [Display, c_int]
libX11.XDefaultVisual.restype = c_void_p

libX11.XDefaultDepth.argtypes = [Display, c_int]
libX11.XDefaultDepth.restype = c_int

libX11.XRootWindow.argtypes = [Display, c_int]
libX11.XRootWindow.restype = Window

libX11.XCreateColormap.restype = c_ulong
libX11.XCreateColormap.argtypes = [Display, Window, c_void_p, c_int]

libX11.XStoreName.argtypes = [Display, Window, c_char_p]
libX11.XStoreName.restype = c_int

libX11.XSelectInput.argtypes = [Display, Window, c_long]
libX11.XSelectInput.restype = c_int

libX11.XMapWindow.argtypes = [Display, Window]
libX11.XMapWindow.restype = c_int

libX11.XPending.argtypes = [Display]
libX11.XPending.restype = c_int

libX11.XGetWindowAttributes.argtypes = [Display, Window, c_void_p]
libX11.XGetWindowAttributes.restype = c_int

libX11.XDestroyWindow.argtypes = [Display, Window]
libX11.XDestroyWindow.restype = c_int

libX11.XFlush.argtypes = [Display]
libX11.XFlush.restype = c_int

libX11.XBlackPixel.argtypes = [Display, c_int]
libX11.XBlackPixel.restype = c_ulong

libX11.XWhitePixel.argtypes = [Display, c_int]
libX11.XWhitePixel.restype = c_ulong

_display = None

class XAnyEvent(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("serial", ctypes.c_ulong),
        ("send_event", Bool),
        ("display", ctypes.c_void_p),
        ("window", Window),
    ]

class XKeyEvent(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("serial", ctypes.c_ulong),
        ("send_event", Bool),
        ("display", ctypes.c_void_p),
        ("window", Window),
        ("root", Window),
        ("subwindow", Window),
        ("time", Time),
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("x_root", ctypes.c_int),
        ("y_root", ctypes.c_int),
        ("state", ctypes.c_uint),
        ("keycode", ctypes.c_uint),
        ("same_screen", Bool),
    ]

class XButtonEvent(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("serial", ctypes.c_ulong),
        ("send_event", Bool),
        ("display", ctypes.c_void_p),
        ("window", Window),
        ("root", Window),
        ("subwindow", Window),
        ("time", Time),
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("x_root", ctypes.c_int),
        ("y_root", ctypes.c_int),
        ("state", ctypes.c_uint),
        ("button", ctypes.c_uint),
        ("same_screen", Bool),
    ]

class XConfigureEvent(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("serial", ctypes.c_ulong),
        ("send_event", Bool),
        ("display", ctypes.c_void_p),
        ("event", Window),
        ("window", Window),
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("width", ctypes.c_int),
        ("height", ctypes.c_int),
        ("border_width", ctypes.c_int),
        ("above", Window),
        ("override_redirect", Bool),
    ]

# Now build the XEvent union
class XEvent(ctypes.Union):
    _fields_ = [
        ("type", ctypes.c_int),
        ("xany", XAnyEvent),
        ("xkey", XKeyEvent),
        ("xbutton", XButtonEvent),
        ("xconfigure", XConfigureEvent),
        ("pad", ctypes.c_byte * 192),  # ensures correct size (192 bytes)
    ]

class XErrorEvent(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int),
        ("display", ctypes.c_void_p),
        ("resourceid", ctypes.c_ulong),
        ("serial", ctypes.c_ulong),
        ("error_code", ctypes.c_ubyte),
        ("request_code", ctypes.c_ubyte),
        ("minor_code", ctypes.c_ubyte),
    ]
    
libX11.XNextEvent.argtypes = [Display, POINTER(XEvent)]
libX11.XNextEvent.restype = c_int

libX11.XCreateColormap.restype = c_ulong
libX11.XCreateColormap.argtypes = [Display, Window, c_void_p, c_int]

libX11.XMatchVisualInfo.argtypes = [Display, c_int, c_int, c_int, POINTER(XVisualInfo)]
libX11.XMatchVisualInfo.restype = c_int

X_EVENT_MASKS = {
    "KEYPRESS":          (1 << 0),
    "KEYRELEASE":        (1 << 1),
    "BUTTONPRESS":       (1 << 2),
    "BUTTONRELEASE":     (1 << 3),
    "ENTERWINDOW":       (1 << 4),
    "LEAVEWINDOW":       (1 << 5),
    "POINTERMOTION":     (1 << 6),
    "FOCUSCHANGE":       (1 << 8),
    "EXPOSURE":          (1 << 15),
    "STRUCTURE":         (1 << 17),
    "PROPERTYCHANGE":    (1 << 19),
    "VISIBILITY":        (1 << 23),
    "SUBSTRUCTURE":      (1 << 25),
    "COLORMAPCHANGE":    (1 << 29)
}

X_EVENT_MASK = (
    X_EVENT_MASKS["KEYPRESS"] |
    X_EVENT_MASKS["KEYRELEASE"] |
    X_EVENT_MASKS["BUTTONPRESS"] |
    X_EVENT_MASKS["BUTTONRELEASE"] |
    X_EVENT_MASKS["ENTERWINDOW"] |
    X_EVENT_MASKS["LEAVEWINDOW"] |
    X_EVENT_MASKS["POINTERMOTION"] |
    X_EVENT_MASKS["FOCUSCHANGE"] |
    X_EVENT_MASKS["EXPOSURE"] |
    X_EVENT_MASKS["STRUCTURE"] |
    X_EVENT_MASKS["PROPERTYCHANGE"] |
    X_EVENT_MASKS["VISIBILITY"] |
    X_EVENT_MASKS["SUBSTRUCTURE"] |
    X_EVENT_MASKS["COLORMAPCHANGE"]
)

CWBackPixel = 1 << 0
CWBorderPixel = 1 << 3
CWEventMask  = 1 << 11
CWColormap = 1 << 13

VALUEMASK = CWBackPixel | CWBorderPixel | CWEventMask | CWColormap

INPUT_OUTPUT = 1

@ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(XErrorEvent))
def x_error_handler(error_event_ptr):
    e = error_event_ptr.contents
    print("X Error caught!")
    print(f"Error code: {e.error_code}")
    print(f"Request code: {e.request_code}")
    print(f"Minor code: {e.minor_code}")
    print(f"Resource ID: {hex(e.resourceid)}")
    print(f"Serial: {e.serial}")
    return 0

libX11.XSetErrorHandler(x_error_handler)

def _get_display():
    global _display
    if not _display:
        libX11.XOpenDisplay.restype = Display
        _display = libX11.XOpenDisplay(None)
        if not _display:
            raise RuntimeError("Unable to Open X11 Display")
    
    return _display

class PIX11Flags:
    def __init__(self, x=100, y=100, width=800, height=600, border=1, border_color=0, background=0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.border = border
        self.border_color = border_color
        self.background = background

def create_window(title: str="Pheonix Ion", width:int=800, height:int=600, flags:PIX11Flags=None):
    display = _get_display()

    screen = libX11.XDefaultScreen(display)
    root = libX11.XRootWindow(display, screen)

    if not root:
        raise RuntimeError("Invalid root window!")

    black = libX11.XBlackPixel(display, screen)
    white = libX11.XWhitePixel(display, screen)

    if not flags:
        flags = PIX11Flags(100, 100, width, height, 1, black, white)

    vinfo = XVisualInfo()
    if not libX11.XMatchVisualInfo(display, screen, 24, 4, ctypes.byref(vinfo)):
        raise RuntimeError("No suitable visual found!")

    visual = ctypes.c_void_p(vinfo.visual)
    depth = vinfo.depth

    attrs = XSetWindowAttributes()
    ctypes.memset(ctypes.byref(attrs), 0, ctypes.sizeof(attrs))
    attrs.background_pixel = flags.background
    attrs.border_pixel = flags.border_color
    attrs.event_mask = X_EVENT_MASK

    colormap = libX11.XCreateColormap(display, root, visual, 0)
    attrs.colormap = colormap

    window = libX11.XCreateWindow(
        display,
        root,
        flags.x, flags.y,
        flags.width, flags.height,
        flags.border,
        depth,
        INPUT_OUTPUT,
        visual,
        VALUEMASK,
        ctypes.byref(attrs)
    )

    if not window:
        raise RuntimeError("Failed to create X11 window")

    # Set title, select input, and map window
    libX11.XStoreName(display, window, title.encode("utf-8"))
    libX11.XSelectInput(display, window, X_EVENT_MASK)

    return (display, window)

def show_window(window_tuple):
    display, window = window_tuple
    libX11.XMapWindow(display, window)
    libX11.XFlush(display)
    
def hide_window(window_tuple):
    display, window = window_tuple
    libX11.XUnmapWindow(display, window)
    libX11.XFlush(display)
    
def set_window_title(window_tuple, title):
    display, window = window_tuple
    libX11.XStoreName(display, window, title.encode("utf-8"))
    
def destroy_window(window_tuple):
    display, window = window_tuple
    libX11.XDestroyWindow(display, window)
    libX11.XFlush(display)

def poll_events(window_tuple):
    display, window = window_tuple

    if not display:
        # No valid display, probably running headless
        return []

    event = XEvent()
    events = []
    
    try:
        while True:
            pending = libX11.XPending(display)
            if pending <= 0:
                break
            libX11.XNextEvent(display, ctypes.byref(event))
            e_type = event.type
            e = event # shortcut

            # --- Keyboard ---
            if e_type == 2:   # KeyPress
                events.append(PIonEvent("KEYDOWN", keycode=e.xkey.keycode, x=e.xkey.x, y=e.xkey.y))
            elif e_type == 3: # KeyRelease
                events.append(PIonEvent("KEYUP", keycode=e.xkey.keycode, x=e.xkey.x, y=e.xkey.y))

            # --- Mouse Buttons ---
            elif e_type == 4: # ButtonPress
                button = e.xbutton.button
                name = {1: "LEFT_DOWN", 2: "MIDDLE_DOWN", 3: "RIGHT_DOWN"}.get(button, "MOUSEDOWN")
                events.append(PIonEvent(name, button=button, x=e.xbutton.x, y=e.xbutton.y))
            elif e_type == 5: # ButtonRelease
                button = e.xbutton.button
                name = {1: "LEFT_UP", 2: "MIDDLE_UP", 3: "RIGHT_UP"}.get(button, "MOUSEUP")
                events.append(PIonEvent(name, button=button, x=e.xbutton.x, y=e.xbutton.y))

            # --- Motion ---
            elif e_type == 6: # MotionNotify
                events.append(PIonEvent("MOUSEMOVE", x=e.xbutton.x, y=e.xbutton.y))

            # --- Window crossing ---
            elif e_type == 7: # EnterNotify
                events.append(PIonEvent("MOUSEENTER", x=e.xbutton.x, y=e.xbutton.y))
            elif e_type == 8: # LeaveNotify
                events.append(PIonEvent("MOUSELEAVE", x=e.xbutton.x, y=e.xbutton.y))

            # --- Focus ---
            elif e_type == 9: # FocusIn
                events.append(PIonEvent("FOCUS_GAINED"))
            elif e_type == 10: # FocusOut
                events.append(PIonEvent("FOCUS_LOST"))

            # --- Expose / Redraw ---
            elif e_type == 12: # Expose
                events.append(PIonEvent("EXPOSE"))

            # --- Visibility ---
            elif e_type == 15: # VisibilityNotify
                events.append(PIonEvent("REDRAW"))

            # --- Structure / Resize / Move ---
            elif e_type == 22: # ConfigureNotify
                events.append(PIonEvent("RESIZE",
                    x=e.xconfigure.x,
                    y=e.xconfigure.y,
                    width=e.xconfigure.width,
                    height=e.xconfigure.height
                ))

            # --- Map / Unmap / Destroy ---
            elif e_type == 19: # MapNotify
                events.append(PIonEvent("SHOW"))
            elif e_type == 18: # UnmapNotify
                events.append(PIonEvent("HIDE"))
            elif e_type == 17: # DestroyNotify
                events.append(PIonEvent("DESTROY", destroyed=False))

            # --- Client / Close requests ---
            elif e_type == 33: # ClientMessage (e.g., WM_DELETE_WINDOW)
                events.append(PIonEvent("CLOSE", destroyed=False))

            # --- Property changed (like window title, icon, etc.) ---
            elif e_type == 28: # PropertyNotify
                events.append(PIonEvent("PROPERTYCHANGE"))

            # --- MappingNotify (keyboard layout changes) ---
            elif e_type == 34:
                events.append(PIonEvent("INPUTLANGCHANGE"))

            else:
                events.append(PIonEvent("UNKNOWN", raw_type=e_type))
    except Exception as e:
        print("X11 Poll Error:", e)
    return events

def is_window_alive(window_tuple) -> bool:
    display, window = window_tuple
    attrs = XWindowAttributes()
    status = libX11.XGetWindowAttributes(display, window, ctypes.byref(attrs))
    return status != 0

def get_gpu(window_tuple, api:Optional[str]=None, driver:Optional[BaseGPUD]=None) -> GPU_API:
    display, window = window_tuple
    gpu = GPU_API(api, driver)
    return gpu

def attach_gpu(window_tuple, gpu:GPU_API) -> None:
    display, window = window_tuple
    gpu.driver.init(display, window, create_and_attach_ctx=True)


"""Vulkan Driver implementation using BaseGPUD class (incomplete)"""

import ctypes
import platform
from typing import Any
from phardwareitk.GPU._base import BaseGPUD

VK_SUCCESS = 0
VK_STRUCTURE_TYPE_APPLICATION_INFO = 0
VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO = 1
VK_API_VERSION_1_0 = (1 << 22) | (0 << 12) | 0 # Vulkan 1.0

class VulkanDriver(BaseGPUD):
    def __init__(self):
        self.platform = platform.system().lower()

        self.lib = None

        if self.platform == "windows":
            self.lib = ctypes.WinDLL("vulkan-1.dll")
        else:
            self.lib = ctypes.CDLL('libvulkan.so.1')

        self.instance = None
        self.physical_device = None
        self.device = None
        self.queue = None
        self.surface = None
        self.swapchain = None
        self.cmd_pool = None

    def init(self, display: Any, window: Any, create_and_attach_ctx = False):
        class VkApplicationInfo(ctypes.Structure):
            _fields_ = [
                ("sType", ctypes.c_int),
                ("pNext", ctypes.c_void_p),
                ("pApplicationName", ctypes.c_char_p),
                ("applicationVersion", ctypes.c_uint32),
                ("pEngineName", ctypes.c_char_p),
                ("engineVersion", ctypes.c_uint32),
                ("apiVersion", ctypes.c_uint32),
            ]

        app_info = VkApplicationInfo(
            sType=VK_STRUCTURE_TYPE_APPLICATION_INFO,
            pNext=None,
            pApplicationName=b"PheonixIon Vulkan",
            applicationVersion=1,
            pEngineName=b"PheonixIon Engine",
            engineVersion=1,
            apiVersion=VK_API_VERSION_1_0,
        )

        class VkInstanceCreateInfo(ctypes.Structure):
            _fields_ = [
                ("sType", ctypes.c_int),
                ("pNext", ctypes.c_void_p),
                ("flags", ctypes.c_uint32),
                ("pApplicationInfo", ctypes.POINTER(VkApplicationInfo)),
                ("enabledLayerCount", ctypes.c_uint32),
                ("ppEnabledLayerNames", ctypes.POINTER(ctypes.c_char_p)),
                ("enabledExtensionCount", ctypes.c_uint32),
                ("ppEnabledExtensionNames", ctypes.POINTER(ctypes.c_char_p)),
            ]

        create_info = VkInstanceCreateInfo(
            sType=VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
            pNext=None,
            flags=0,
            pApplicationInfo=ctypes.pointer(app_info),
            enabledLayerCount=0,
            ppEnabledLayerNames=None,
            enabledExtensionCount=0,
            ppEnabledExtensionNames=None,
        )

        vkCreateInstance = self.load_function("vkCreateInstance")
        vkCreateInstance.restype = ctypes.c_int
        vkCreateInstance.argtypes = [ctypes.POINTER(VkInstanceCreateInfo), ctypes.c_void_p, ctypes.POINTER(ctypes.c_void_p)]

        instance = ctypes.c_void_p()
        result = vkCreateInstance(ctypes.byref(create_info), None, ctypes.byref(instance))
        if result != VK_SUCCESS:
            raise RuntimeError(f"Failed to create Vulkan instance: {result}")

        self.instance = instance

        if create_and_attach_ctx:
            self.create_context(display, window)

    def load_function(self, name: str):
        try:
            return getattr(self.lib, name)
        except AttributeError:
            return None
        
    def create_context(self, display: Any, window: Any):
        pass

    def shutdown(self):
        if self.instance:
            vkDestroyInstance = self.load_function("vkDestroyInstance")
            if vkDestroyInstance:
                vkDestroyInstance(self.instance, None)
            self.instance = None



import platform
import os

platform = platform.system().lower()

Windows = True if platform == "windows" else False
Linux = True if platform == "linux" else False
Kali = True if platform == "kali" else False
MacOs = True if platform == "darwin" else False
Ubuntu = True if platform == "ubuntu" else False


import os
import platform
import json

mem = None
vga = None
framebuffer = None
disk_data = None
sys_data = None

posix_os = [
    "linux",
    "unix",
    "euleros",
    "macos",
    "openbsd",
    "plan9",
    "vxworks",
    "minix",
    "ubuntu",
    "debian",
    "fedora",
    "arch",
    "aix",
    "hp-ux",
    "solaris",
    "netbsd",
    "freebsd",
    "dragonflybsd",
    "qnx",
    "illumos",
    "truenas",
    "redhat",
    "centos",
    "rocky linux",
    "almalinux",
    "suse",
    "gentoo",
    "slackware",
    "mandriva",
    "scientific linux",
    "oracle linux",
]

class PenvInturruptReturn(Exception):
    def __init__(self, *values):
        self.values = values
        super().__init__(*values)

    def __str__(self):
        val = ""
        for value in self.values:
            val += str(value) + " "
        return val

    def __repr__(self):
        return self.__str__()

def set_mem(memory: object) -> None:
    global mem
    if memory == None:
        return
    mem = memory

def set_vga(vga_: object) -> None:
    global vga
    if vga_ == None:
        return
    vga = vga_

def set_framebuffer(framebuffer_: object) -> None:
    global framebuffer
    if framebuffer_ == None:
        return
    framebuffer = framebuffer_

def set_disk_data(disk_data_: object) -> None:
    global disk_data
    if disk_data_ == None:
        return
    disk_data = disk_data_

def set_sys_data(sys_data_: object) -> None:
    global sys_data
    if sys_data_ == None:
        return
    sys_data = sys_data_

def interrupt(code: int, *args):
    global mem
    global vga
    global framebuffer
    global sys_data
    global disk_data

    if code == 0x0:  # Write
        print(*args)
    elif code == 0x1:  # Read
        return input(args[0])
    elif code == 0x2:  # Clear
        if platform.system().lower() == "windows":
            os.system("cls")
        elif platform.system().lower() in posix_os:
            os.system("cls")
        else:
            print("\033[2J\033[H")

    elif code == 0x51:  # Shift to PMode
        table: dict = args[0]
        table["type"] = 0x51
        return_ = json.dumps(table)
        raise PenvInturruptReturn(return_)  # Stop the exec function
    elif code == 0x60:  # Call file and exit
        table: dict = {"type": 0x60, "filename": args[0]}
        return_ = json.dumps(table)
        raise PenvInturruptReturn(return_)

    elif code == 0x4:  # Get Ram
        if not mem:
            return None
        else:
            return mem.get_ram(args[0], args[1])
    elif code == 0x5:  # Get VGA
        if not vga:
            return None
        else:
            return vga.serialize()
    elif code == 0x6:  # Frambuffer Write
        if not framebuffer:
            return None
        else:
            return framebuffer.write_pixel(args[0], args[1], args[2], args[3], args[4])
    elif code == 0x7:  # VGA Write
        if not vga:
            return None
        else:
            return vga.write_pixel(args[0], args[1], args[2], args[3], args[4])
    elif code == 0x8:  # Get Disk Data
        return disk_data
    elif code == 0x9:  # Get Sys Data
        return sys_data
    elif code == 0x10:  # Write mem func
        if not mem:
            return None
        else:
            return mem.write_ram(args[0], args[1], args[2])
    elif code == 0x11:  # Get Ram
        if not mem:
            return None
        else:
            ram = mem.ram
            mem.ram = None
            return ram
    elif code == 0x12:  # Return RAM
        mem.ram = args[0]
    elif code == 0x13:  # Framebuffer Clear
        if not framebuffer:
            return None
        else:
            return framebuffer.clear()
    elif code == 0x14:  # Framebuffer Render
        if not framebuffer:
            return None
        else:
            return framebuffer.render()

def ddh_key(dict_: dict, key: str) -> bool:
    try:
        val = dict_[key]
        return True
    except Exception:
        return False

def get_os() -> tuple[bool, bool, bool, str]:
	os_ = platform.system().lower()
	win32: bool = False
	posix: bool = False
	unknown: bool = True

	posix_os = [
		"linux",
		"unix",
		"euleros",
		"macos",
		"openbsd",
		"plan9",
		"vxworks",
		"minix",
		"ubuntu",
		"debian",
		"fedora",
		"arch",
		"aix",
		"hp-ux",
		"solaris",
		"netbsd",
		"freebsd",
		"dragonflybsd",
		"qnx",
		"illumos",
		"truenas",
		"redhat",
		"centos",
		"rocky linux",
		"almalinux",
		"suse",
		"gentoo",
		"slackware",
		"mandriva",
		"scientific linux",
		"oracle linux",
	]

	if os_ == "windows":
		win32 = True
		unknown = False
		posix = False
	elif os_ in posix_os:
		posix = True
		unknown = False
		win32 = False
	else:
		unknown = True
		win32 = False
		posix = False

	return (win32, posix, unknown, os_)


# NEVER GIVE UP!!!
import os
import sys
import platform

module_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "..")
if not module_path in sys.path:
    sys.path.append(module_path)

if platform.system().lower() == "windows":
    from ctypes import wintypes
    import ctypes
    class WindowsFrameBufferDriver:
        def __init__(self, w, h, d, classn="Static", windown="Pheonix Virtual Environment"):
            from phardwareitk.GUI import gui_sdl as gui

            self.w:int = w
            self.h:int = h
            self.d:int = d

            self.pitch = (self.w * self.d + 3) & ~3
            self.vram = bytearray(self.pitch * self.h)

            self.classn = classn
            self.windown = windown

            self.deleted = False

            self.gui = gui

            self.gui.initialize()
            self.gui.InitializePError()
            self.gui.ExtensionsInit()
            self.window = self.gui.CreateWindow(windown, w, h, self.gui.SDL_WINDOWPOS_CENTERED, None, 0, 0)
            self.renderer = self.gui.MakeRenderer(self.window, -1, self.gui.SDL_RENDERER_ACCELERATED)

            self.gui.UpdateWindow(self.window)

            self.gui.TitleBarWindows(None, "dark")

            self.gui.UpdateWindow(self.window)

        def wp(self, x, y, r, g, b):
            if 0 <= x < self.w and 0 <= y < self.h:
                offset = (y * self.pitch + x) * self.d
                self.vram[offset:offset+3] = bytes([r, g, b])

        def cl(self, r, g, b):
            color = bytes([r, g, b])
            for y in range(self.h):
                for x in range(self.w):
                    offset = (y * self.pitch + x) * self.d
                    self.vram[offset:offset+3] = color

        def flush(self):
            vram_ptr = (ctypes.c_ubyte * len(self.vram)).from_buffer(self.vram)
            surface = self.gui.SDL_CreateRGBSurfaceFrom(vram_ptr, self.w, self.h, 24, self.pitch,
                                                   0x000000FF, 0x0000FF00, 0x00FF0000, 0)
            self.gui.SDL_BlitSurface(surface, None, self.gui.SDL_GetWindowSurface(self.window), None)
            self.gui.SDL_UpdateWindowSurface(self.window)

        def msg_loop(self):
            self.gui.EventLoop(True, self.window, None, None, self.renderer, True)
            self.__del__()

        def __del__(self):
            if not self.deleted:
                self.gui.DestroyWindow(self.window, None, None)
                self.gui.DeinitializePError()
                self.gui.Quit()
                self.deleted = True

else:
    import mmap
    class LinuxFrameBufferDriver:
        def __init__(self, w, h, d):
            self.w = w
            self.h = h
            self.d = d

            self.fb = os.open("/dev/fb0", os.O_RDWR)
            self.screensize = self.w * self.h * self.d

            self.vram = bytearray(self.screensize)
            self.mm = mmap.mmap(
                self.fb, self.screensize, mmap.MAP_SHARED, mmap.PROT_WRITE | mmap.PROT_READ
            )

        def wp(self, x, y, r, g, b):
            if x < 0 or (y < 0 or (x >= self.w or y >= self.h)):
                return

            offset = (y * self.w + x) * self.d
            self.vram[offset : offset + 3] = bytes([b, g, r])  # Linux has BGR framebuffer

        def cl(self, r, g, b):
            for y in range(self.h):
                for x in range(self.w):
                    self.wp(x, y, r, g, b)

        def flush(self):
            self.mm.seek(0)
            self.mm.write(self.vram)

        def __del__(self):
            self.mm.close()
            os.close(self.fb)



"""C Standard Int"""

from . import *

int8_t = char
int16_t = short
int32_t = Int
int64_t = long
uint8_t = unsigned_char
uint16_t = unsigned_short
uint32_t = unsigned_int
uint64_t = unsigned_long

Int8_t = int8_t
Int16_t = int16_t
Int32_t = int32_t
Int64_t = int64_t
Uint8_t = uint8_t
Uint16_t = uint16_t
Uint32_t = uint32_t
Uint64_t = uint64_t



"""GPU level access"""

import subprocess
import platform
import importlib

def get_gpu_type():
    try:
        # Check for NVIDIA GPU using nvidia-smi command
        if platform.system() == "Linux":
            try:
                result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if result.returncode == 0:
                    return "NVIDIA"
            except Exception:
                pass
        elif platform.system() == "Windows":
            try:
                result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if result.returncode == 0:
                    return "NVIDIA"
            except Exception:
                pass

        # Check for Intel UHD Graphics
        if platform.system() == "Windows":
            try:
                result = subprocess.run(['wmic', 'path', 'win32_VideoController', 'get', 'name'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if b'Intel' in result.stdout:
                    return "Intel UHD"
            except Exception:
                pass

        # Check for AMD GPU
        if platform.system() == "Windows":
            try:
                result = subprocess.run(['wmic', 'path', 'win32_VideoController', 'get', 'name'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if b'AMD' in result.stdout:
                    return "AMD"
            except Exception:
                pass

        # For Linux, we can check the /proc directory
        if platform.system() == "Linux":
            try:
                with open('/proc/driver/nvidia/version', 'r') as f:
                    if 'NVIDIA' in f.read():
                        return "NVIDIA"
                with open('/var/log/Xorg.0.log', 'r') as f:
                    log_content = f.read()
                    if 'Intel' in log_content:
                        return "Intel UHD"
                    elif 'AMD' in log_content:
                        return "AMD"
            except Exception:
                pass

    except Exception as e:
        print(f"Error detecting GPU: {e}")

    return "No compatible GPU detected"

gpu_type = None

def set_gpu_type():
    # Import the appropriate module based on the GPU type
    gpu_type = get_gpu_type()
    if gpu_type == "NVIDIA":
        importlib.import_module("GPU.NVIDIA")
    elif gpu_type == "Intel UHD":
        importlib.import_module("GPU.IntelUHD")
    elif gpu_type == "AMD":
        importlib.import_module("GPU.AMD")
    else:
        print("Using CPU or unsupported GPU")


import os
import sys
import struct
from . import *

if not module_path in sys.path:
    sys.path.append(module_path)

if not os.path.dirname(__file__) in sys.path:
    sys.path.append(os.path.dirname(__file__))

from PBFS import PBFS_HEADER

from phardwareitk.Memory import Memory as Mem
from phardwareitk.PENV.shared import *
from phardwareitk.PENV.framebuffer import Framebuffer
from phardwareitk.PENV import PBFS

SystemData: dict = {}
drive_data: dict = {}

DRIVE_PATH = os.path.join(base_path, "PheonixSSD.pbfs")
DRIVE_FOLDER = os.path.join(base_path, ".hard_disk")

# Get System Info
def GetSystemInfoBIOS() -> None:
    """Gets System Info"""
    import psutil

    SystemData["cpu_cores"] = os.cpu_count()
    SystemData["cpu_architecture"] = platform.machine()
    SystemData["processor_name"] = platform.processor()
    SystemData["host_name"] = platform.node()
    SystemData["host_system"] = platform.system()
    SystemData["host_version"] = platform.version()
    SystemData["host_release"] = platform.release()
    SystemData["host_platform"] = platform.platform()
    SystemData["host_uname"] = platform.uname()
    SystemData["ram_size"] = round(
        psutil.virtual_Mem().total / (1024**3), 2
    )  # in GB
    SystemData["storage_size"] = round(
        psutil.disk_usage("/").total / (1024**3), 2
    )  # in GB

def get_drive_data() -> int:
    if not os.path.exists(DRIVE_PATH):
        return -1  # Drive Error

    if not os.path.exists(os.path.join(DRIVE_PATH, "vdrive.pbfs")):
        return -2  # Drive Not Found

    with open(os.path.join(DRIVE_PATH, "vdrive.pbfs"), "rb") as f:
        data = f.read()

    drive_data = struct.unpack(PBFS_HEADER_FORMAT, data[:68])

    drive_data_ = {
        "magic": drive_data[0].decode("utf-8"),
        "block_size": drive_data[1],
        "total_blocks": drive_data[2],
        "disk_name": drive_data[3].decode("utf-8"),
        "timestamp": drive_data[4],
        "version": drive_data[5],
        "first_boot_timestamp": drive_data[6],
        "os_boot_mode": drive_data[7],
        "uuid": data[68:].hex()
    }

    return drive_data_

def search_for_bootloader() -> str:
    """Searches for the bootloader in the drive data directory."""
    # search the directory for file -> boot.info
    bootinfo_path = os.path.join(DRIVE_FOLDER, "boot.info")
    bootinfo_found: bool = True
    bootloader_path = ""
    if not os.path.exists(bootinfo_path):
        bootinfo_found = False

    # IF oot Info file is not found, it means this is a first time boot
    if not bootinfo_found:
        # Search the dir for all .py files
        files = []

        for file_ in os.listdir(DRIVE_FOLDER):
            if file_.endswith(".py") or file_.endswith(".vasm"):
                files.append(file_)

        if not files:
            # Search all 512-byte files
            for file_ in os.listdir(DRIVE_FOLDER):
                if os.stat(os.path.join(DRIVE_FOLDER, file_)).st_size == 512:
                    files.append(file_)

        print("Acceptable maybe bootloader files:", files)

        # Get the last 4 bytes of the file
        bootloader_path = ""
        buff: bytes = b""
        print("Trying to find bootloader signature (AA55)")
        for file_ in files:
            with open(os.path.join(DRIVE_FOLDER, file_), "rb") as f:
                f.seek(-4, os.SEEK_END)
                buff = f.read(4)
                if buff == b"AA55":
                    bootloader_path = os.path.join(DRIVE_FOLDER, file_)
                if buff == b"A55\n": #Incase of auto newline
                    f.seek(-5, os.SEEK_END)
                    buff = f.read(4)
                    if buff == b"AA55":
                        bootloader_path = os.path.join(DRIVE_FOLDER, file_)

        # Make bootinfo file
        if not bootloader_path:
            print("No bootloader found. Please install a bootloader.")
            return ""
        else:
            print("Writing boot.info ...")
            with open(bootinfo_path, "wb") as f:
                data = bootloader_path + "\n" + platform.system().lower()
                f.write(data.encode("utf-8"))
    else:
        with open(bootinfo_path, "rb") as f:
            bootloader_path = f.read().decode("utf-8").split("\n")
            required_os = bootloader_path[1]
            bootloader_path = bootloader_path[0]

        if not required_os == platform.system().lower():
            print("Boot.Info file was created in another OS!")
            os.remove(bootinfo_path)
            bootloader_path = search_for_bootloader()

        if not os.path.exists(bootloader_path):
            print("Boot.Info file doesn't contain a valid path, rewriting...")
            os.remove(bootinfo_path)
            bootinfo_path = search_for_bootloader()

    return bootloader_path

def run_bootloader(file_path: str, mem: Mem.Memory, bios_ram_start_addr: int, fb: Framebuffer) -> tuple:
    set_mem(mem)

    size_of_bl: int = os.stat(file_path).st_size

    if size_of_bl > 512:
        return ("!ERROR!", None, None, None)

    with open(file_path, "rb") as f:
        f.seek(0)
        mem.write_ram(f.read(), bios_ram_start_addr)

    set_framebuffer(fb)
    _drive_data = get_drive_data()
    set_disk_data(_drive_data)
    set_sys_data(GetSystemInfoBIOS())

    exec_globals = {}
    exec_globals["__int__"] = interrupt

    print("Running Bootloader...")
    try:
        exec(mem.get_ram(size_of_bl, bios_ram_start_addr).decode("utf-8"), exec_globals)
    except PenvInturruptReturn as e:
        if _drive_data["first_boot_timestamp"] == 0:
            import time
            import uuid
            # Write the timestamp to the drive data
            drive_data = []
            drive_data.append(_drive_data["magic"].encode("utf-8"))
            drive_data.append(_drive_data["block_size"])
            drive_data.append(_drive_data["total_blocks"])
            drive_data.append(_drive_data["disk_name"].encode("utf-8"))
            drive_data.append(_drive_data["timestamp"])
            drive_data.append(_drive_data["version"])
            drive_data.append(int(time.time()))
            drive_data.append(_drive_data["os_boot_mode"])
            drive_data.append(0)
            drive_data.append(0)

            data = struct.pack(PBFS_HEADER_FORMAT, *drive_data)
            data += uuid.UUID(bytes=bytes.fromhex(_drive_data["uuid"])).bytes

            with open(os.path.join(DRIVE_PATH, "vdrive.pbfs"), "wb") as f:
                f.write(data)

            print("Updated first boot timestamp to:", int(time.time()))

        return (str(e), SystemData, _drive_data, DRIVE_PATH)

def start() -> str:
    GetSystemInfoBIOS()
    get_drive_data()
    file_ = search_for_bootloader()
    set_sys_data(SystemData)
    set_disk_data(disk_data)
    return file_



"""This file includes everything you can think of for making C/Python/Asm/C++/C#/etc code for windows, which will run on any OS that python runs on."""

import sys
from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions.C import *

from typing import * # type: ignore

# Winuser.h -> Class Styles
CS_BYTEALIGNCLIENT = 0x1000
CS_BYTEALIGNWINDOW = 0x2000
CS_CLASSDC = 0x0040
CS_DBLCLKS = 0x0008
CS_DROPSHADOW = 0x00020000
CS_GLOBALCLASS = 0x4000
CS_HREDRAW = 0x0002
CS_NOCLOSE = 0x0200
CS_OWNDC = 0x0020
CS_PARENTDC = 0x0080
CS_SAVEBITS = 0x0800
CS_VREDRAW = 0x0001

# winuser.h -> Icons
IDI_APPLICATION = 32512
IDI_ERROR = 32513
IDI_QUESTION = 32514
IDI_WARNING = 32515
IDI_INFORMATION = 32516
IDI_WINLOGO = 32517
IDI_SHIELD = 32518

# winuser.h -> Cursors
IDC_ARROW = 32512
IDC_IBEAM = 32513
IDC_WAIT = 32514
IDC_CROSS = 32515
IDC_UPARROW = 32516
IDC_SIZENWSE = 32642
IDC_SIZENESW = 32643
IDC_SIZEWE = 32644
IDC_SIZENS = 32645
IDC_SIZEALL = 32646
IDC_NO = 32648
IDC_HAND = 32649
IDC_APPSTARTING = 32650
IDC_HELP = 32651
IDC_PIN = 32671
IDC_PERSON = 32672

# wingdi -> GetStockObject
# These are copied from win32con (only these ones not others)
WHITE_BRUSH = 0
LTGRAY_BRUSH = 1
GRAY_BRUSH = 2
DKGRAY_BRUSH = 3
BLACK_BRUSH = 4
NULL_BRUSH = 5
HOLLOW_BRUSH = NULL_BRUSH
WHITE_PEN = 6
BLACK_PEN = 7
NULL_PEN = 8
OEM_FIXED_FONT = 10
ANSI_FIXED_FONT = 11
ANSI_VAR_FONT = 12
SYSTEM_FONT = 13
DEVICE_DEFAULT_FONT = 14
DEFAULT_PALETTE = 15

# winuser.h -> Size and location
CW_USEDEFAULT = -2147483648

# winuser.h -> Show Window
SW_HIDE = 0
SW_SHOWNORMAL = 1
SW_NORMAL = 1
SW_SHOWMINIMIZED = 2
SW_SHOWMAXIMIZED = 3
SW_MAXIMIZED = 3
SW_SHOWNOACTIVATE = 4
SW_SHOW = 5
SW_MINIMIZE = 6
SW_SHOWMINNOACTIVE = 7
SW_SHOWNA = 8
SW_RESTORE = 9
SW_SHOWDEFAULT = 10
SW_FORCEMINIMIZE = 11

# wingdi -> Painting and Drawing Messages
WM_DISPLAYCHANGE = 126
WM_ERASEBKGND = 20
WM_NCPAINT = 133
WM_PAINT = 15
WM_PRINT = 791
WM_PRINTCLIENT = 792
WM_SETREDRAW = 11
WM_SYNCPAINT = 136
# Windows Core Messages
WM_DESTROY = 2
WM_CLOSE = 16

# winuser.h -> PeekMessage Constants Part-1
PM_NOREMOVE = 0x0000
PM_REMOVE = 0x0001
PM_NOYIELD = 0x0002

# Basic Classes and typedefs

class LPCWSTR(str):
    """Represents a pointer to a wide-character string (Unicode)."""

class DWORD(int):
    """Represents a 32-bit unsigned integer."""

class HANDLE(int):
    """Represents a handle to an object (file, process, etc.)."""
    INVALID_HANDLE_VALUE = -1  # Represents a failure case

class BOOL(int):
    """Represents a bool.
    """

class BYTE(int):
  """Represents a byte."""
  pass

class WORD(int):
  """Represents a word (16-bit unsigned integer)"""
  pass

class LONG(int):
    """Represents a LONG (32-bit signed integer)."""
    pass

class PLONG(Pointer):
    """Represents a pointer to a LONG (32-bit signed integer)."""

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError("PLONG must point to an int.")
        super().__init__()

class LPVOID(Pointer):
    """A Pointer to a void type."""

    def __init__(self, value: bytearray):
        if not isinstance(value, (bytes, bytearray)):
            raise TypeError("LPVOID must point to bytes or bytearray.")
        super().__init__(value)

class LPCVOID(Pointer):
    """A Pointer to a constant void type (immutable raw bytes)."""

    def __init__(self, value: bytes):
        if not isinstance(value, bytes):
            raise TypeError("LPCVOID must point to immutable bytes.")
        super().__init__(value)

class LPDWORD(Pointer):
    """A Pointer to a DWORD (32-bit unsigned integer)."""

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError("LPDWORD must point to an int.")
        super().__init__(value)

class LPLONG(Pointer):
    """A Pointer to a LONG (32-bit signed integer)."""

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError("LPLONG must point to an int.")
        super().__init__(value)

class LPOVERLAPPED(Pointer):
    """A Pointer to an OVERLAPPED structure."""

    def __init__(self, value: dict):
        if not isinstance(value, dict):
            raise TypeError("LPOVERLAPPED must point to a dictionary.")
        super().__init__(value)

class LPSECURITY_ATTRIBUTES(Pointer):
    """A Pointer to security attributes."""

    def __init__(self, value: dict):
        if not isinstance(value, dict):
            raise TypeError("LPSECURITY_ATTRIBUTES must point to a dictionary.")
        super().__init__(value)

# File System and IO
# fileapi, winioctl, ioapiset, namedpipeapi, pipe

class FileAccessMode:
    """A class for all the File Access modes, example -> read, write, etc
    """
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    GENERIC_EXECUTE = 0x20000000
    ALL = 0x10000000

    @staticmethod
    def convert_to_python(value: Union[int, DWORD]) -> str:
        """
        Converts a DWORD value to a Python file mode string (e.g., "r", "w", "w+", "a+").
        """
        if not isinstance(value, int):
            return ""

        read = bool(value & FileAccessMode.GENERIC_READ)
        write = bool(value & FileAccessMode.GENERIC_WRITE)
        execute = bool(value & FileAccessMode.GENERIC_EXECUTE)
        all_access = bool(value & FileAccessMode.GENERIC_ALL)

        if all_access:
            return "r+b"  # Example of all access. Adjust if necessary.

        if read and write:
            return "r+" if not execute else "r+b"  # Binary if execute is true.
        if read:
            return "r" if not execute else "rb"  # Binary if execute is true.
        if write:
            return "w" if not execute else "wb"  # Binary if execute is true.
        if execute:
            return "x"  # or "xb". Adjust as needed.

        return ""

class FileDisposition:
    """Defines Windows file creation/opening behaviors."""

    CREATE_NEW = 1
    CREATE_ALWAYS = 2
    OPEN_EXISTING = 3
    OPEN_ALWAYS = 4
    TRUNCATE_EXISTING = 5

class FilePointerMoveMethod:
    """Defines how file pointers are moved."""

    FILE_BEGIN = 0
    FILE_CURRENT = 1
    FILE_END = 2

class FileHandleManager:
    """Manages file handles."""

    _handle_counter = 4  # Windows starts at small non-zero values.
    _open_files: Dict[int, IO] = {}

    @classmethod
    def create_handle(cls, file_path: str, mode: str) -> HANDLE:
        """Creates and registers a file handle."""
        try:
            file_obj = open(file_path, mode)
            handle_id = cls._handle_counter
            cls._open_files[handle_id] = file_obj
            cls._handle_counter += 1
            return HANDLE(handle_id)
        except IOError:
            return HANDLE(HANDLE.INVALID_HANDLE_VALUE)

    @classmethod
    def get_file(cls, handle: HANDLE) -> Optional[IO]:
        """Retrieves a file object from a handle."""
        return cls._open_files.get(handle)

    @classmethod
    def close_handle(cls, handle: HANDLE) -> BOOL:
        """Closes a handle and removes it from tracking."""
        file_obj = cls._open_files.pop(handle, None)
        if file_obj:
            file_obj.close()
            return BOOL(True)
        return BOOL(False)

class FILE_TYPE:
    """File types returned by GetFileType"""
    UNKNOWN = 0x0000
    DISK = 0x0001
    CHAR = 0x0002
    PIPE = 0x0003

class FILE_INFO:
    """Represents file information returned by GetFileInformationByHandle."""
    def __init__(self, file_size: DWORD, creation_time: DWORD, last_access_time: DWORD, last_write_time: DWORD):
        self.file_size = file_size
        self.creation_time = creation_time
        self.last_access_time = last_access_time
        self.last_write_time = last_write_time

class FileAttributes:
    """Represents file attribute flags used in Windows API."""

    READONLY = 0x00000001
    HIDDEN = 0x00000002
    SYSTEM = 0x00000004
    DIRECTORY = 0x00000010
    ARCHIVE = 0x00000020
    DEVICE = 0x00000040
    NORMAL = 0x00000080
    TEMPORARY = 0x00000100
    SPARSE_FILE = 0x00000200
    REPARSE_POINT = 0x00000400
    COMPRESSED = 0x00000800
    OFFLINE = 0x00001000
    NOT_CONTENT_INDEXED = 0x00002000
    ENCRYPTED = 0x00004000

class FileAPI:
    @staticmethod
    def CreateFileW(lpFileName:LPCWSTR, dwDesiredAccess:Union[FileAccessMode, DWORD], dwShareMode:DWORD, lpSecurityAttributes:LPSECURITY_ATTRIBUTES, dwCreationDisposition:DWORD, dwFlagsAndAttributes:DWORD, hTemplateFile:HANDLE) -> HANDLE:
        """
        Opens or creates a file, device, pipe, or other system object.

        Parameters:
            lpFileName (LPCWSTR): The name of the file or device to be created/opened.
            dwDesiredAccess (FileAccessMode | DWORD): The requested access (read, write, execute).
            dwShareMode (DWORD): The sharing mode (e.g., read/write sharing).
            lpSecurityAttributes (Optional[LPSECURITY_ATTRIBUTES]): Security attributes for the file.
            dwCreationDisposition (DWORD): Action to take if the file exists or not (CREATE_NEW, OPEN_EXISTING, etc.).
            dwFlagsAndAttributes (DWORD): File attributes and flags (e.g., FILE_ATTRIBUTE_NORMAL).
            hTemplateFile (HANDLE): Handle to a template file (if applicable).

        Returns:
            HANDLE: A handle to the created or opened file. INVALID_HANDLE_VALUE (-1) on failure.
        """
        mode = {
            FileDisposition.CREATE_NEW: "x",
            FileDisposition.CREATE_ALWAYS: "w",
            FileDisposition.OPEN_EXISTING: "r",
            FileDisposition.OPEN_ALWAYS: "a+",
            FileDisposition.TRUNCATE_EXISTING: "w"
        }.get(dwCreationDisposition, "r")

        # Convert Windows FileAccess flags into Python file modes
        if dwDesiredAccess & FileAccessMode.GENERIC_WRITE:
            mode += "+"
        elif dwDesiredAccess & FileAccessMode.GENERIC_READ:
            mode = "r"

        return FileHandleManager.create_handle(lpFileName, mode)

    @staticmethod
    def CloseHandle(hObject:HANDLE) -> BOOL:
        """Closes a handle."""
        return FileHandleManager.close_handle(hObject)

    @staticmethod
    def ReadFile(hFile: HANDLE, lpBuffer: LPVOID, nNumberOfBytesToRead: DWORD, lpNumberOfBytesRead: LPDWORD, lpOverlapped: LPOVERLAPPED) -> BOOL:
        """
        Reads data from the specified file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            lpBuffer (LPVOID): Buffer to receive data.
            nNumberOfBytesToRead (DWORD): Maximum number of bytes to read.
            lpNumberOfBytesRead (LPDWORD): Pointer to variable that receives the number of bytes read.
            lpOverlapped (LPOVERLAPPED): Pointer to an OVERLAPPED structure (for async I/O, not implemented).

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            data = file.read(nNumberOfBytesToRead)
            lpBuffer[:len(data)] = data.encode()  # Store data in lpBuffer
            lpNumberOfBytesRead.value = len(data)  # Store bytes read
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def WriteFile(hFile: HANDLE, lpBuffer: LPCVOID, nNumberOfBytesToWrite: DWORD, lpNumberOfBytesWritten: LPDWORD, lpOverlapped: LPOVERLAPPED) -> BOOL:
        """
        Writes data to the specified file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            lpBuffer (LPCVOID): The data to write.
            nNumberOfBytesToWrite (DWORD): Number of bytes to write.
            lpNumberOfBytesWritten (LPDWORD): Pointer to variable that receives the number of bytes written.
            lpOverlapped (LPOVERLAPPED): Pointer to an OVERLAPPED structure (for async I/O, not implemented).

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            written = file.write(lpBuffer[:nNumberOfBytesToWrite].decode())  # Write bytes
            file.flush()  # Ensure data is saved
            lpNumberOfBytesWritten.value = written  # Store bytes written
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def SetFilePointer(hFile: HANDLE, lDistanceToMove: LONG, lpDistanceToMoveHigh: PLONG, dwMoveMethod: DWORD) -> DWORD:
        """
        Moves the file pointer of an open file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            lDistanceToMove (LONG): Number of bytes to move.
            lpDistanceToMoveHigh (PLONG): Pointer to high-order bytes of move distance (not used in this implementation).
            dwMoveMethod (DWORD): Move method (beginning, current, or end).

        Returns:
            DWORD: The new file pointer position, or -1 on failure.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return DWORD(-1)

        try:
            import os
            if dwMoveMethod == FilePointerMoveMethod.FILE_BEGIN:
                file.seek(lDistanceToMove, os.SEEK_SET)
            elif dwMoveMethod == FilePointerMoveMethod.FILE_CURRENT:
                file.seek(lDistanceToMove, os.SEEK_CUR)
            elif dwMoveMethod == FilePointerMoveMethod.FILE_END:
                file.seek(lDistanceToMove, os.SEEK_END)

            return DWORD(file.tell())  # Return new file pointer position
        except Exception:
            return DWORD(-1)

    @staticmethod
    def GetFileSize(hFile:HANDLE, lpFileSizeHigh:LPDWORD) -> DWORD:
        """
        Retrieves the size of a specified file.

        Parameters:
            hFile (HANDLE): Handle to the file whose size is being queried.
            lpFileSizeHigh (LPDWORD): Pointer to a variable that receives the high-order DWORD of the file size.

        Returns:
            DWORD: The low-order part of the file size. Returns INVALID_FILE_SIZE (-1) on failure.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return DWORD(-1)  # INVALID_FILE_SIZE

        try:
            import os
            current_pos = file.tell()  # Save current position
            file.seek(0, os.SEEK_END)
            size = file.tell()
            file.seek(current_pos, os.SEEK_SET)  # Restore file position

            if lpFileSizeHigh is not None:
                lpFileSizeHigh.value = (size >> 32) & 0xFFFFFFFF  # Set high-order 32 bits

            return DWORD(size & 0xFFFFFFFF)  # Return low-order 32 bits
        except Exception:
            return DWORD(-1)

    @staticmethod
    def FlushFileBuffers(hFile: HANDLE) -> BOOL:
        """
        Flushes buffers of the specified file, ensuring data is written to disk.

        Parameters:
            hFile (HANDLE): Handle to the file.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            file.flush()
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def SetEndOfFile(hFile: HANDLE) -> BOOL:
        """
        Truncates or extends a file to the current file pointer position.

        Parameters:
            hFile (HANDLE): Handle to the file.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            file.truncate()
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def GetFileInformationByHandle(hFile: HANDLE) -> Optional[FILE_INFO]:
        """
        Retrieves file metadata.

        Parameters:
            hFile (HANDLE): Handle to the file.

        Returns:
            Optional[FILE_INFO]: File metadata if successful, otherwise None.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return None

        try:
            import os

            stats = os.stat(file.fileno())
            return FILE_INFO(
                file_size=DWORD(stats.st_size),
                creation_time=DWORD(int(stats.st_ctime)),
                last_access_time=DWORD(int(stats.st_atime)),
                last_write_time=DWORD(int(stats.st_mtime)),
            )
        except Exception:
            return None

    @staticmethod
    def GetFileType(hFile: HANDLE) -> DWORD:
        """
        Retrieves the type of file (disk, char, or pipe).

        Parameters:
            hFile (HANDLE): Handle to the file.

        Returns:
            DWORD: File type.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return DWORD(FILE_TYPE.UNKNOWN)

        if file.isatty():
            return DWORD(FILE_TYPE.CHAR)
        return DWORD(FILE_TYPE.DISK)

    @staticmethod
    def DuplicateHandle(hSourceHandle: HANDLE) -> HANDLE:
        """
        Duplicates an existing handle.

        Parameters:
            hSourceHandle (HANDLE): Handle to duplicate.

        Returns:
            HANDLE: A new handle if successful, INVALID_HANDLE_VALUE (-1) otherwise.
        """
        file = FileHandleManager.get_file(hSourceHandle)
        if not file:
            return HANDLE(HANDLE.INVALID_HANDLE_VALUE)

        try:
            new_handle = FileHandleManager.create_handle(file.name, file.mode)
            return new_handle
        except Exception:
            return HANDLE(HANDLE.INVALID_HANDLE_VALUE)

    @staticmethod
    def LockFile(hFile: HANDLE, dwFileOffsetLow: DWORD, dwFileOffsetHigh: DWORD, nNumberOfBytesToLock: DWORD) -> BOOL:
        """
        Locks a region of a file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            dwFileOffsetLow (DWORD): Low-order part of the starting byte offset.
            dwFileOffsetHigh (DWORD): High-order part of the starting byte offset.
            nNumberOfBytesToLock (DWORD): Number of bytes to lock.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        # Simulating a lock (Python does not have direct lock API)
        return BOOL(True)

    @staticmethod
    def UnlockFile(hFile: HANDLE, dwFileOffsetLow: DWORD, dwFileOffsetHigh: DWORD, nNumberOfBytesToUnlock: DWORD) -> BOOL:
        """
        Unlocks a previously locked region.

        Parameters:
            hFile (HANDLE): Handle to the file.
            dwFileOffsetLow (DWORD): Low-order part of the starting byte offset.
            dwFileOffsetHigh (DWORD): High-order part of the starting byte offset.
            nNumberOfBytesToUnlock (DWORD): Number of bytes to unlock.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        # Simulating unlock (Python does not have direct unlock API)
        return BOOL(True)

    @staticmethod
    def LockFileEx(hFile: HANDLE, dwFlags: DWORD, dwReserved: DWORD, nNumberOfBytesToLockLow: DWORD, nNumberOfBytesToLockHigh: DWORD, lpOverlapped: LPOVERLAPPED) -> BOOL:
        """
        Locks a region of a file with additional options.

        Parameters:
            hFile (HANDLE): Handle to the file.
            dwFlags (DWORD): Locking flags.
            dwReserved (DWORD): Reserved, must be zero.
            nNumberOfBytesToLockLow (DWORD): Low-order part of the lock range.
            nNumberOfBytesToLockHigh (DWORD): High-order part of the lock range.
            lpOverlapped (LPOVERLAPPED): Overlapped structure.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        return FileAPI.LockFile(hFile, nNumberOfBytesToLockLow, nNumberOfBytesToLockHigh, 0)

    @staticmethod
    def UnlockFileEx(hFile: HANDLE, dwReserved: DWORD, nNumberOfBytesToUnlockLow: DWORD, nNumberOfBytesToUnlockHigh: DWORD, lpOverlapped: LPOVERLAPPED) -> BOOL:
        """
        Unlocks a previously locked region (extended).

        Parameters:
            hFile (HANDLE): Handle to the file.
            dwReserved (DWORD): Reserved, must be zero.
            nNumberOfBytesToUnlockLow (DWORD): Low-order part of the unlock range.
            nNumberOfBytesToUnlockHigh (DWORD): High-order part of the unlock range.
            lpOverlapped (LPOVERLAPPED): Overlapped structure.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        return FileAPI.UnlockFile(hFile, nNumberOfBytesToUnlockLow, nNumberOfBytesToUnlockHigh, 0)

    @staticmethod
    def DeleteFileW(lpFileName:LPCWSTR) -> BOOL:
        """
        Deletes a File.

        Parameters:
            lpFileName (LPCWSTR): The name of the file to be deleted.

        Returns:
            BOOL: True if successfull, else False
        """

        import os

        if not os.path.exists(lpFileName):
            return BOOL(False)

        try:
            os.remove(lpFileName)
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def GetFileAttributesW(lpFileName:LPCWSTR) -> DWORD:
        """Gets the file Attributes

        Args:
            lpFileName (LPCWSTR): The name of the file or directory.

        Returns:
            DWORD: If the function succeeds, the return value contains the attributes of the specified file or directory, If the function fails, the return value is **INVALID_FILE_ATTRIBUTES**.
        """
        import os
        import stat

        if not os.path.exists(lpFileName):
            return DWORD(-1)

        try:
            attributes = 0

            #Assign attribs
            if os.path.isdir(lpFileName):
                attributes |= FileAttributes.DIRECTORY
            if os.path.isfile(lpFileName):
                attributes |= FileAttributes.NORMAL
            if not os.access(lpFileName, os.W_OK):
                attributes |= FileAttributes.READONLY
            if os.stat(lpFileName).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN:
                attributes |= FileAttributes.HIDDEN

            return DWORD(attributes)
        except Exception:
            return DWORD(-1)

    @staticmethod
    def SetFileAttributesW(lpFileName:LPCWSTR, dwFileAttributes: DWORD) -> BOOL:
        """
        Sets attributes for the specified file.

        Parameters:
            lpFileName (LPCWSTR): The name of the file.
            dwFileAttributes (DWORD): The new file attributes to be set.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        import os
        import stat

        if not os.path.exists(lpFileName):
            return BOOL(False)

        try:
            # Remove READONLY FLAG, if needed
            if dwFileAttributes & FileAttributes.READONLY:
                os.chmod(lpFileName, stat.S_IREAD)
            else:
                os.chmod(lpFileName, stat.S_IWRITE)

            # Handle hidden attribute (cross-platform)
            file_dir, file_name = os.path.split(lpFileName)
            new_file_name = f".{file_name}" if dwFileAttributes & FileAttributes.HIDDEN and not file_name.startswith(".") else file_name
            new_file_path = os.path.join(file_dir, new_file_name)

            # Rename only if necessary
            if new_file_path != lpFileName:
                os.rename(lpFileName, new_file_path)

            return BOOL(True)

            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def GetFileTime(hFile: HANDLE, lpCreationTime:LPDWORD, lpLastAccessTime:LPDWORD, lpLastWriteTime: LPDWORD) -> BOOL:
        """
        Retrieves the creation, last access, and last write times of a file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            lpCreationTime (LPDWORD): Pointer to store creation time.
            lpLastAccessTime (LPDWORD): Pointer to store last access time.
            lpLastWriteTime (LPDWORD): Pointer to store last write time.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        import os
        import time

        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            stats = os.stat(file.name)
            lpCreationTime.value = DWORD(int(stats.st_birthtime))
            lpLastAccessTime.value = DWORD(int(stats.st_atime))
            lpLastWriteTime.value = DWORD(int(stats.st_mtime))
            return BOOL(True)
        except Exception:
            return BOOL(False)

    @staticmethod
    def SetFileTime(hFile: HANDLE, lpCreationTime: LPDWORD, lpLastAccessTime: LPDWORD, lpLastWriteTime: LPDWORD) -> BOOL:
        """
        Sets the creation, last access, and last write times of a file.

        Parameters:
            hFile (HANDLE): Handle to the file.
            lpCreationTime (LPDWORD): Pointer to new creation time.
            lpLastAccessTime (LPDWORD): Pointer to new last access time.
            lpLastWriteTime (LPDWORD): Pointer to new last write time.

        Returns:
            BOOL: True if successful, False otherwise.
        """
        import os
        import time

        file = FileHandleManager.get_file(hFile)
        if not file:
            return BOOL(False)

        try:
            times = (
                lpLastAccessTime.value if lpLastAccessTime else None,
                lpLastWriteTime.value if lpLastWriteTime else None,
            )
            os.utime(file.name, times)
            return BOOL(True)
        except Exception:
            return BOOL(False)




"""Provides many extensions"""

from typing import *

import os
import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

class Color:
    """Color Class for defining color.
    """
    def __init__(self, color:str="white", alpha:int=255) -> None:
        """Init of color class.

        Args:
            color (str, optional): The color, leave and chose a function if, want to use RGB. Defaults to "white".
        """
        self.color_names_rgb:dict[str, tuple[int, int, int]] = {
            "red": (255, 0, 0),
            "green": (0, 255, 0),
            "blue": (0, 0, 255),
            "yellow": (255, 255, 0),
            "cyan": (0, 255, 255),
            "magenta": (255, 0, 255),
            "black": (0, 0, 0),
            "white": (255, 255, 255),
            "gray": (169, 169, 169),
            "grey": (169, 169, 169),
            "silver": (192, 192, 192),
            "maroon": (128, 0, 0),
            "olive": (128, 128, 0),
            "lime": (0, 255, 0),
            "aqua": (0, 255, 255),
            "fuchsia": (255, 0, 255),
            "teal": (0, 128, 128),
            "navy": (0, 0, 128),
            "purple": (128, 0, 128),
            "pink": (255, 192, 203),
            "brown": (139, 69, 19),
            "beige": (245, 245, 220),
            "chocolate": (210, 105, 30),
            "coral": (255, 127, 80),
            "indigo": (75, 0, 130),
            "lavender": (230, 230, 250),
            "violet": (238, 130, 238),
            "peach": (255, 218, 185),
            "gold": (255, 215, 0),
            "plum": (221, 160, 221),
            "orchid": (218, 112, 214),
            "crimson": (220, 20, 60),
            "chartreuse": (127, 255, 0),
            "tomato": (255, 99, 71),
            "salmon": (250, 128, 114),
            "seashell": (255, 245, 238),
            "blush": (222, 93, 131),
            "wheat": (245, 222, 179),
            "khaki": (240, 230, 140),
            "mint": (189, 252, 201),
            "forest": (34, 139, 34),
            "lemon": (255, 247, 0),
            "pearl": (234, 234, 234),
            "periwinkle": (204, 204, 255),
            "sand": (194, 178, 128),
            "emerald": (80, 200, 120),
            "russet": (128, 70, 27),
            "scarlet": (255, 36, 0),
            "charcoal": (54, 69, 79),
            "rose": (255, 99, 123),
            "tangerine": (255, 129, 0),
            "bordeaux": (127, 0, 24),
            "light-cyan-special": (0, 200, 255),
            "light-red": (255, 102, 102),
            "light-green": (144, 238, 144),
            "light-blue": (173, 216, 230),
            "light-yellow": (255, 255, 153),
            "light-cyan": (224, 255, 255),
            "light-magenta": (255, 182, 193),
            "light-gray": (211, 211, 211),
            "light-orange": (255, 200, 124),
            "light-brown": (181, 101, 29),
            "light-purple": (216, 191, 216),
            "light-pink": (255, 182, 193),
            "light-gold": (255, 236, 139),
            "light-silver": (230, 230, 230),
            "light-maroon": (205, 92, 92),
            "light-olive": (204, 255, 102),
            "light-lime": (204, 255, 204),
            "light-aqua": (204, 255, 255),
            "light-fuchsia": (255, 153, 255),
            "light-teal": (153, 255, 204),
            "light-navy": (102, 153, 255),
            "light-indigo": (180, 160, 255),
            "light-lavender": (230, 230, 250),
            "light-violet": (238, 210, 255),
            "light-peach": (255, 218, 185),
            "light-plum": (233, 190, 255),
            "light-orchid": (218, 182, 214),
            "light-crimson": (255, 102, 140),
            "light-chartreuse": (204, 255, 102),
            "light-tomato": (255, 160, 122),
            "light-salmon": (255, 160, 122),
            "light-seashell": (255, 245, 238),
            "light-blush": (255, 178, 185),
            "light-wheat": (250, 240, 190),
            "light-khaki": (240, 230, 140),
            "light-mint": (189, 252, 201),
            "light-forest": (144, 238, 144),
            "light-lemon": (255, 255, 153),
            "light-pearl": (242, 242, 242),
            "light-sand": (238, 214, 175),
            "light-emerald": (160, 255, 160),
            "light-rose": (255, 192, 203),
            "light-tangerine": (255, 204, 153),
            "light-bordeaux": (205, 92, 120),
            "light-coral": (240, 128, 128),
            "light-charcoal": (180, 180, 180),
            "light-sky": (176, 226, 255),
            "light-mocha": (210, 180, 140),
            "light-coffee": (200, 170, 120),
            "light-cream": (255, 253, 208),
            "bronze": (205, 127, 50),
            "copper": (184, 115, 51),
            "brass": (181, 166, 66),
            "iron": (70, 76, 79),
            "steel": (70, 130, 180),
            "gunmetal": (42, 52, 57),
            "platinum": (229, 228, 226),
            "titanium": (135, 134, 129),
            "chrome": (200, 200, 210),
            "pewter": (139, 139, 139),
            "nickel": (176, 176, 176),
            "aluminum": (212, 212, 212),
            "lead": (90, 90, 90),
            "zinc": (188, 201, 194),
            "cobalt": (0, 71, 171),
            "mud": (96, 70, 38),
            "clay": (177, 115, 75),
            "dirt": (155, 118, 83),
            "soil": (134, 95, 60),
            "stone": (140, 130, 120),
            "rock": (105, 95, 85),
            "ash": (178, 190, 181),
            "dust": (209, 182, 156),
            "smoke": (115, 130, 118),
            "char": (70, 65, 60),
            "bark": (92, 64, 51),
            "moss": (138, 154, 91),
            "fern": (79, 121, 66),
            "pine": (1, 121, 111),
            "sage": (158, 169, 132),
            "olive-drab": (107, 142, 35),
            "sandstone": (236, 213, 164),
            "topsoil": (101, 67, 33),
            "mahogany": (192, 64, 0),
            "espresso": (60, 44, 33),
            "coffee": (111, 78, 55),
            "mocha": (150, 102, 69),
            "caramel": (255, 190, 111),
            "butterscotch": (255, 204, 102),
            "tan": (210, 180, 140),
            "sepia": (112, 66, 20),
            "skyblue": (135, 206, 235),
            "deepsky": (0, 191, 255),
            "midnight": (25, 25, 112),
            "ocean": (0, 105, 148),
            "seafoam": (159, 226, 191),
            "tealblue": (54, 117, 136),
            "aquamarine": (127, 255, 212),
            "turquoise": (64, 224, 208),
            "cerulean": (42, 82, 190),
            "azure": (240, 255, 255),
            "arctic": (176, 224, 230),
            "ice": (153, 255, 255),
            "wave": (66, 146, 198),
            "storm": (80, 88, 99),
            "mist": (224, 240, 255),
            "fog": (215, 215, 215),
            "glacier": (175, 238, 238),
            "ember": (255, 89, 0),
            "lava": (207, 16, 32),
            "fire": (255, 85, 0),
            "flame": (255, 153, 51),
            "burnt-orange": (204, 85, 0),
            "sunset": (250, 128, 114),
            "sunrise": (255, 204, 153),
            "amber": (255, 191, 0),
            "honey": (234, 179, 0),
            "goldenrod": (218, 165, 32),
            "mustard": (255, 219, 88),
            "candlelight": (255, 255, 153),
            "torch": (255, 140, 0),
            "slate": (112, 128, 144),
            "granite": (112, 110, 102),
            "graphite": (64, 64, 64),
            "obsidian": (15, 15, 15),
            "onyx": (53, 56, 57),
            "pebble": (176, 170, 160),
            "cement": (200, 200, 200),
            "concrete": (169, 169, 169),
            "smokestack": (110, 110, 120),
            "neon-pink": (255, 20, 147),
            "neon-blue": (77, 77, 255),
            "neon-green": (57, 255, 20),
            "neon-orange": (255, 153, 51),
            "neon-yellow": (255, 255, 102),
            "neon-purple": (204, 0, 255),
            "neon-cyan": (0, 255, 255),
            "neon-lime": (191, 255, 0),
            "neon-red": (255, 0, 64),
            "dark-red": (139, 0, 0),
            "dark-green": (0, 100, 0),
            "dark-blue": (0, 0, 139),
            "dark-cyan": (0, 139, 139),
            "dark-magenta": (139, 0, 139),
            "dark-gray": (64, 64, 64),
            "dark-orange": (255, 140, 0),
            "dark-brown": (101, 67, 33),
            "dark-purple": (90, 0, 120),
            "dark-yellow": (204, 204, 0),
            "dark-pink": (231, 84, 128),
            "dark-turquoise": (0, 206, 209),
            "dark-teal": (0, 128, 128),
            "dark-olive": (85, 107, 47),
            "dark-gold": (184, 134, 11),
            "dark-crimson": (139, 0, 0),
            "pastel-blue": (174, 198, 207),
            "pastel-green": (119, 221, 119),
            "pastel-yellow": (253, 253, 150),
            "pastel-pink": (255, 179, 186),
            "pastel-orange": (255, 203, 164),
            "pastel-purple": (207, 159, 255),
            "pastel-cyan": (178, 255, 255),
            "pastel-lavender": (220, 208, 255),
            "pastel-lime": (210, 255, 210),
            "pastel-gray": (230, 230, 230),
            "grass": (124, 252, 0),
            "leaf": (50, 205, 50),
            "bamboo": (181, 201, 142),
            "jungle": (29, 97, 49),
            "mossy-green": (102, 153, 102),
            "ivy": (85, 107, 47),
            "fern-green": (79, 121, 66),
            "pine-green": (1, 121, 111),
            "apple-green": (141, 182, 0),
            "forest-night": (34, 78, 34),
            "midnight-dream": (25, 25, 112),
            "galaxy": (25, 25, 112),
            "space": (20, 24, 82),
            "starlight": (230, 230, 255),
            "moonstone": (214, 236, 241),
            "nebula": (120, 81, 169),
            "cosmic": (44, 0, 88),
            "aether": (200, 255, 255),
            "void": (10, 10, 30),
            "plasma": (255, 51, 153),
            "infrared": (255, 36, 0),
            "ultraviolet": (128, 0, 255),
            "stormcloud": (100, 100, 120),
            "twilight": (54, 47, 98),
            "shadow": (48, 47, 47),
            "phantom": (50, 50, 60),
            "specter": (170, 170, 200),
            "ivory": (255, 255, 240),
            "cream": (255, 253, 208),
            "bone": (227, 218, 201),
            "linen": (250, 240, 230),
            "almond": (239, 222, 205),
            "champagne": (247, 231, 206),
            "sand-dune": (237, 201, 175),
            "vanilla": (243, 229, 171),
            "latte": (230, 190, 138),
            "hazel": (218, 145, 0),
            "terminal-green": (0, 255, 64),
            "terminal-blue": (0, 128, 255),
            "terminal-yellow": (255, 255, 128),
            "error-red": (255, 64, 64),
            "warning-orange": (255, 128, 0),
            "success-green": (0, 255, 128),
            "info-blue": (0, 191, 255),
            "highlight": (255, 255, 160),
            "accent": (255, 128, 192),
            "ruby": (224, 17, 95),
            "sapphire": (15, 82, 186),
            "emerald-green": (80, 200, 120),
            "jade": (0, 168, 107),
            "amethyst": (153, 102, 204),
            "topaz": (255, 200, 124),
            "opal": (168, 195, 188),
            "garnet": (115, 54, 53),
            "rose-quartz": (255, 182, 193),
            "onyx-black": (15, 15, 15),
        }

        self.hex_codes_rgb:dict[str, tuple[int, int, int]] = {
            f"#{r:02x}{g:02x}{b:02x}": (r, g, b)
            for name, (r, g, b) in self.color_names_rgb.items()
        }

        self.color_names_hex:dict[str, str] = {
            name: f"#{r:02x}{g:02x}{b:02x}"
            for name, (r, g, b) in self.color_names_rgb.items()
        }

        self.color = self.ColorToRGB(color.lower())
        self.alpha = alpha

    def RGB(self, r:int, g:int, b:int) -> tuple:
        """Sets the color to the specified RGB values

        Args:
            r (int): Red.
            g (int): Green.
            b (int): Blue.

        Returns:
            tuple: Just a copy of the entered data in the form of a tuple.
        """
        self.color = (r, g, b)
        return (r, g, b)

    def ColorToRGB(self, color:str) -> tuple:
        """Converts the specified color into its specified RGB values IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            color (str): The color name.

        Returns:
            tuple: The RGB values.
        """
        if not color.lower() in self.color_names_rgb.keys():
            return None

        return self.color_names_rgb[color.lower()]

    def HexToRGB(self, hex_code:str) -> tuple:
        """Converts Hex code to RGB values IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            hex_code (str): The hexadecimal code.

        Returns:
            tuple: The RGB values.
        """
        if not hex_code.lower() in self.hex_codes_rgb.keys():
            return None

        return self.hex_codes_rgb[hex_code.lower()]

    def ColorToHex(self, color:str) -> tuple:
        """Converts a color name to a hexadecimal code IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            color (str): The color name.

        Returns:
            tuple: The Hexadecimal value.
        """
        if not color.lower() in self.color_names_hex.keys():
            return None

        return self.color_names_hex[color.lower()]

    def to_rgb_code(self):
        """Converts the RGB tuple to an ANSI escape code."""
        return f"\033[38;2;{self.color[0]};{self.color[1]};{self.color[2]}m"

    def to_background_code(self):
        """Converts the RGB tuple to an ANSI background escape code."""
        return f"\033[48;2;{self.color[0]};{self.color[1]};{self.color[2]}m"

class PHeader:
    """Generates a PHardware Header.

    Args:
        version (int, optional): The version of the program/File. Defaults to 1.
        flags (int, optional): The flags.. Defaults to 0.
        magicCode (Union[str, bytes], optional): The magic code in ASCII or string. Defaults to "PHardwareITK".
    """
    def __init__(self, version:int=1, flags:int=0, magicCode:Union[str, bytes]="PHardwareITK") -> None:
        """Initialize the PHeader.

        Args:
            version (int, optional): The version of the program/File. Defaults to 1.
            flags (int, optional): The flags.. Defaults to 0.
            magicCode (Union[str, bytes], optional): The magic code in ASCII or string. Defaults to "PHardwareITK".
        """
        self.version = version
        self.flags = flags
        self.magicCode = magicCode
        self.checksum = None

        if isinstance(magicCode, str):
            self.magicCode = magicCode.encode('ascii')
            magicCode = magicCode.encode('ascii')
        elif isinstance(magicCode, bytes):
            try:
                magicCode.decode('ascii')
            except UnicodeDecodeError:
                return None

    def calculateChecksum(self, Header:bytes) -> str:
        """Calculate checksum of Header.

        Args:
            Header (bytes): The header.

        Returns:
            str: The checksum.
        """
        import hashlib
        return hashlib.sha256(Header).hexdigest()

    def Pack(self) -> Any:
        """Packs the specified data into a single PHeader.


        The structure includes:


        - Magic number (4 bytes)

        - Version (4 bytes)

        - Flags (4 bytes)

        - Placeholder for checksum (32 bytes as hexadecimal string)


        Returns:
            Any: The header.
        """
        import struct

        headerData = struct.pack('>4sIB', self.magicCode, self.version, self.flags)

        checkSum = self.calculateChecksum(headerData)

        return headerData + bytes.fromhex(checkSum)

    def UnPack(self, data:str) -> Any:
        """Upacks the binary header into header fields.
        """
        import struct
        header = data[:-64]
        checksum = data[-64:]

        magic, version, flags = struct.unpack('>4sIB', header)

        self.magic = magic
        self.version = version
        self.flags = flags
        self.checksum = checksum.decode()

        # Validate checksum
        expected_checksum = self.calculateChecksum(header)
        if self.checksum != expected_checksum:
            raise ValueError(f"Checksum mismatch: expected {expected_checksum}, but got {self.checksum}")

        return self

    def WriteToFile(self, filename):
        """
        Write the packed header data to a binary file.
        """
        header_data = self.Pack()
        with open(filename, 'wb') as f:
            f.write(header_data)

    def ReadFromFile(self, filename):
        """
        Read the header from a binary file and unpack the fields.
        """
        if not os.path.exists(filename):
            raise FileNotFoundError(f"{filename} does not exist.")

        with open(filename, 'rb') as f:
            data = f.read()

        return self.UnPack(data)

class PBin:
    """
    PBin (PHardware Binary) is a custom binary serialization format designed to serialize various data types into a
    compact binary representation. The format includes strings, integers, floats, and custom flags.

    Attributes:
        data (bytes): The binary data representing the serialized object.
    """
    def __init__(self, header:Optional[PHeader]) -> None:
        """
        Initializes the PBin object with an empty data field.
        """
        self.data:bytes = b""

        self.header:PHeader = header if header else PHeader()

    def Serialize(self, data:Union[str, bytes, int, dict, float, list, bool], CB_flag:Optional[bool]=None) -> bytes:
        """
        Serializes a string, integer, float, and a custom flag into a binary format.

        Args:
            data (Union[str, bytes (UTF-8), int, dict (CONVERTED TO STR), float, list, bool]): The data to serialize, in UTF-8 format.
            CB_flag (Optional[bool], optional): A flag indicating whether a custom byte follows. Defaults to None

        Returns:
            bytes: The serialized binary data.
        """
        import random

        if isinstance(data, list):
            data = "".join(data)
        elif isinstance(data, bytes):
            data = data.decode()
        elif isinstance(data, int) or (isinstance(data, float) or (isinstance(data, dict) or (isinstance(data, bool)))):
            data = str(data)

        binary_data = bytearray()

        # Serialize the header first using the PHeader class
        header_data = self.header.Pack()
        binary_data.extend(header_data)

        # 1. String serialization (length-prefixed UTF-8 string)
        encoded_string: bytes = data.encode('utf-8')
        binary_data.append(len(encoded_string))  # Length of the string as a single byte.
        binary_data.extend(encoded_string)  # Actual string data

        # 2. Custom flag section (1 byte flag to indicate custom data)
        if CB_flag:
            binary_data.append(0xFF)  # Custom flag byte to indicate extra data
            binary_data.append(random.randint(0, 255))  # Random byte between 0 and 255 for custom data

        # Return the complete binary data.
        self.data = bytes(binary_data)
        return self.data

    def Deserialize(self, binary_data:bytes) -> tuple[str, bool, Optional[int]]:
        """
        Deserializes the provided binary data into its original components.

        Args:
            binary_data (bytes): The binary data to deserialize.

        Returns:
            tuple: A tuple containing the deserialized values in the following order:
                - string_data (str): The deserialized string.
                - CB_flag (bool): Flag indicating whether custom data exists.
                - CB_data (Optional[int]): Custom data byte if present.
        """
        import io
        import struct

        # Unpack the header first
        header_data = binary_data[:self.header.Pack().__len__()]
        self.header.UnPack(header_data)

        # Continue with the rest of the deserialization
        stream = io.BytesIO(binary_data[len(header_data):])

        # 1. Deserialize string (length-prefixed)
        string_length: int = ord(stream.read(1))  # Read the length of the string (1 byte)
        string_data: str = stream.read(string_length).decode('utf-8')  # Read the string

        # 2. Check if there's a custom data flag
        custom_flag: bool = False
        custom_data: int = None
        if stream.read(1) == b'\xFF':
            custom_flag = True
            custom_data = ord(stream.read(1))  # Read the custom data byte

        return (string_data, custom_flag, custom_data)

    @property
    def BinaryData(self):
        return self.data

def ToBytes(data:Optional[Union[str, dict, int, bool]]) -> Optional[bytes]:
    """Encode the given data.

    Args:
        data (Optional[Union[str, dict, int, bool]]): The data.

    Returns:
        Optional[bytes]: Encoded data.
    """
    if not data:
        return None
    data = str(data)
    return data.encode()

def FromBytes(data:Optional[bytes]) -> Optional[str]:
    """Decode the given data.

    Args:
        data (Optional[bytes]): The data.

    Returns:
        Optional[str]: Decoded data.
    """
    if not data:
        return None
    return data.decode()

class TextFont:
    """Just a Class for Font.
    """
    def __init__(self, font:str="Arial", font_size:int=11, font_color:Color=Color(), font_background_color:Color=Color("black"), Italic:bool=False, Underline:bool=False,
                 Bold:bool=False, XtraBold:bool=False, StrikeThrough:bool=False) -> None:
        """Initialize for TextFont class.

        Args:
            font (str, optional): Font. Defaults to "Arial".
            font_size (int, optional): The size of the font. Defaults to 11.
            font_color (Color, optional): The color of the font. Defaults to Color().
            font_background_color (Color, optional): The bg color of the font. Defaults to Color("black").
            Italic (bool, optional): Font Italic. Defaults to False.
            Underline (bool, optional): Underlined Font. Defaults to False.
            Bold (bool, optional): Bold Font. Defaults to False.
            XtraBold (bool, optional): More Bolder Font. Defaults to False.
            StrikeThrough (bool, optional): StikeThrough Font. Defaults to False.
        """

        if Bold and XtraBold:
            Bold = False

        self.font = font
        self.size = font_size
        self.color = font_color
        self.background_color = font_background_color
        self.italic = Italic
        self.underline = Underline
        self.bold = Bold
        self.xtraBold = XtraBold
        self.strike_through = StrikeThrough

    def to_font_code(self):
        """Generate the font style escape codes."""
        codes = []
        if self.bold:
            codes.append("1")  # Bold
        if self.xtraBold:
            codes.append("8")  # Extra Bold (if supported by terminal)
        if self.italic:
            codes.append("3")  # Italic
        if self.underline:
            codes.append("4")  # Underline
        if self.strike_through:
            codes.append("9")  # Strike-through

        if codes:
            return f"\033[{';'.join(codes)}m"
        return ""

    def to_reset_code(self):
        """Reset the font styles."""
        return "\033[0m"

    def to_font_size_code(self):
        """Returns the Escape codes for Font Size. NOTE: Does't work with all OS/Terminals. Due to Font size not being supported universally.
        OS like -> MacOS, XTerm Based, GNOME (Not via Escape codes), Windows (By changing data in a windows file), etc do support it.

        Supported Platforms for this function -> MacOS (ITerm2), Xterm Based (Includes Konsole).
        """

        if self.size == 11:
            return None # Default size

        if "iTerm" in os.environ.get("TERM_PROGRAM", ""):
            return f"\033]50;SetProfile=font-size={self.size}\a" #ITerm2
        elif "xterm" in os.environ.get("TERM", "") or "konsole" in os.environ.get("TERM", ""):
            return f"ESC]50;SetProfile=font-size={self.size}" # Xterm & Konsole
        else:
            return None # Other OS/Terminals

class PWidget:
    def __init__(self, x:int=0, y:int=0, width:int=5, height:int=5, radius:int=0, bgColor:Color=Color(), textWrap:bool=False, textFont:TextFont=TextFont(), text:str="") -> None:
        """Initialize function for PWidget class.

        Args:
            x (int, optional): X coordinate. Defaults to 0.
            y (int, optional): Y coordinate. Defaults to 0.
            width (int, optional): Width of widget. Defaults to 5.
            height (int, optional): Height of widget. Defaults to 5.
            radius (int, optional): Radius of widget (If required). Defaults to 0.
            bgColor (Color, optional): Background color of widget. Defaults to Color().
            textWrap (bool, optional): Text In Widget?. Defaults to False.
            textFont (TextFont, optional): The Text Font (if required). Defaults to TextFont().
            text (str, optional): The text. (if required). Defaults to "".
        """
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.radius = radius
        self.bgColor = bgColor
        self.textWrap = textWrap
        self.textFont = textFont
        self.text = text

    def Contains(self, x:int, y:int) -> bool:
        """Check if a point (x, y) is within the widget's boundaries."""
        return self.x <= x <= self.x + self.width and self.y <= y <= self.y + self.height

    def Draw(self, window: Any):
        """Draw the widget onto the window. Should be overridden by subclasses."""
        raise NotImplementedError("draw() method must be implemented by the subclass.")

    def Change(self, x:Optional[int]=None, y:Optional[int]=None, width:Optional[int]=None, height:Optional[int]=None, radius:Optional[int]=None, bgColor:Optional[Color]=None, textWrap:Optional[bool]=None, textFont:Optional[TextFont]=None, text:Optional[str]=None) -> None:
        """Change the details.

        Args:
            x (Optional[int], optional): X. Defaults to None.
            y (Optional[int], optional): Y. Defaults to None.
            width (Optional[int], optional): Width. Defaults to None.
            height (Optional[int], optional): Height. Defaults to None.
            radius (Optional[int], optional): Radius. Defaults to None.
            bgColor (Optional[Color], optional): Background Color. Defaults to None.
            textWrap (Optional[bool], optional): Text Wrap?. Defaults to None.
            textFont (TextFont, optional): Text Font. Defaults to None.
            text (Optional[str], optional): Text. Defaults to None.
        """
        if x:
            self.x = x

        if y:
            self.y = y

        if width:
            self.width = width

        if height:
            self.height = height

        if radius:
            self.radius = radius

        if bgColor:
            self.bgColor = bgColor

        if textWrap != None:
            self.textWrap = textWrap

        if text:
            self.text = text

        if textFont:
            self.textFont = textFont

class PIcon:
    """PIcon (PHardware Icon) class.
    """
    def __init__(self, iconPath: Optional[str] = "", x: Optional[int] = None, y: Optional[int] = None, width: Optional[int] = None, height: Optional[int] = None) -> None:
        """The INIT function of the PIcon class.

        Args:
            iconPath (Optional[str], optional): The path to the icon file. Defaults to "".
            x (Optional[int], optional): The X coordinate of the icon. Defaults to None.
            y (Optional[int], optional): The Y coordinate of the icon. Defaults to None.
            width (Optional[int], optional): The Width of the icon. Defaults to None.
            height (Optional[int], optional): The Height of the icon. Defaults to None.
        """
        self.iconPath:Optional[str] = iconPath
        self.BiconPath:Optional[bytes] = self.ToBytes(iconPath)
        self.iconSurface:Optional[Any] = None
        self.x:Optional[int] = x
        self.y:Optional[int] = y
        self.width:Optional[int] = width
        self.height:Optional[int] = height

    def LoadImageSDL2(self) -> Any:
        """Loads the image for SDL2.

        Returns:
            Any: Loaded image.
        """
        import sdl2.sdlimage
        return sdl2.sdlimage.IMG_Load(self.BiconPath)

    def SetIconSDL2(self, window:Any) -> None:
        """Changes the icon in the specified window. SDL2.

        Args:
            window (Any): The window.
        """
        import sdl2
        import sdl2.sdlimage

        if self.BiconPath:
            self.iconSurface = self.LoadImageSDL2()
            if not self.iconSurface:
                return None

            sdl2.SDL_SetWindowIcon(window, self.iconSurface)
        else:
            return None

    def Free(self) -> None:
        """Frees the image.
        """
        import sdl2
        if self.iconSurface:
            sdl2.SDL_FreeSurface(self.iconSurface)

    def ToBytes(self, data:Optional[Union[str, dict, int, bool]]) -> Optional[bytes]:
        """Encode the given data.

        Args:
            data (Optional[Union[str, dict, int, bool]]): The data.

        Returns:
            Optional[bytes]: Encoded data.
        """
        if not data:
            return None
        data = str(data)
        return data.encode()

    def FromBytes(self, data:Optional[bytes]) -> Optional[str]:
        """Decode the given data.

        Args:
            data (Optional[bytes]): The data.

        Returns:
            Optional[str]: Decoded data.
        """
        if not data:
            return None
        return data.decode()



"""C-Spectre is a ctypes like library offering auto-resolving, object-oriented, simpler-interface, library-auto-loading, auto-function-resolving and more features while being fully compatible with ctypes"""
import ctypes
import os
import sys
from typing import *

#macros
NULL = 0x0

#types
class CInt:
    """Base Int class providing all functions"""
    def __init__(self, value:int=0):
        self.value = self._wrap()

    def __int__(self):
        return self.value

    def __repr__(self):
        return str(self.value)

    def _wrap(self, value):
        if self.signed:
            mask = 2**(8*self.size) - 1
            value &= mask
            if value > self.max_value:
                value -= (mask + 1)
            return value
        else:
            return value & self.max_value

    def __add__(self, val:int):
        return type(self)(self.value + int(val))

    def __sub__(self, val:int):
        return type(self)(self.value - int(val))

    def __mul__(self, val:int):
        return type(self)(self.value * int(val))

    def __div__(self, val:int):
        return type(self)(self.value / int(val))

    def to_ctypes(self):
        """Returns CTypes compatible instance."""
        if self.size == 1:
            if self.signed:
                return ctypes.c_int8(self.value)
            else:
                return ctypes.c_uint8(self.value)
        elif self.size == 2:
            if self.signed:
                return ctypes.c_int16(self.value)
            else:
                return ctypes.c_uint16(self.value)
        elif self.size == 4:
            if self.signed:
                return ctypes.c_int32(self.value)
            else:
                return ctypes.c_uint32(self.value)
        elif self.size == 8:
            if self.signed:
                return ctypes.c_int64(self.value)
            else:
                return ctypes.c_uint64(self.value)
        else:
            raise ValueError(f"No such C Type (INT) with size -> {self.size}")

class CUint(CInt):
    """Base Unsigned int class providing all functions"""
    def __init__(self, value:int):
        self.signed = True
        super().__init__(value)

class CUint8(CUint):
    """Unsigned Int8 (uint8_t)"""
    def __init__(self, value:int):
        self.size = 1
        self.min_value = 0x0
        self.max_value = 0xFF
        super().__init__(value)

class CUint16(CUint):
    """Unsigned int16 (uint16_t)"""
    def __init__(self, value:int):
        self.size = 2
        self.min_value = 0x0
        self.max_value = 0xFFFF
        super().__init__(value)

class CUint32(CUint):
    """Unsigned int32 (uint32_t)"""
    def __init__(self, value:int):
        self.size = 4
        self.min_value = 0x0
        self.max_value = 0xFFFFFFFF
        super().__init__(value)

class CUint64(CUint):
    """Unsigned int64 (uint64_t)"""
    def __init__(self, value:int):
        self.size = 8
        self.min_value = 0x0
        self.max_value = 0xFFFFFFFFFFFFFFFF
        super().__init__(value)

class CInt8(CInt):
    """Int8 (int8_t)"""
    def __init__(self, value:int):
        self.size = 1
        self.min_value = 0x7F
        self.max_value = 0x80
        super().__init__(value)

class CInt16(CInt):
    """Int16 (int16_t)"""
    def __init__(self, value:int):
        self.size = 2
        self.min_value = 0x7FFF
        self.max_value = 0x8000
        super().__init__(value)

class CInt32(CInt):
    """Int32 (int32_t)"""
    def __init__(self, value:int):
        self.size = 4
        self.min_value = 0x7FFFFFFF
        self.max_value = 0x80000000
        super().__init__(value)

class CInt64(CInt):
    """Int64 (int64_t)"""
    def __init__(self, value:int):
        self.size = 8
        self.min_value = 0x7FFFFFFFFFFFFFFF
        self.max_value = 0x8000000000000000
        super().__init__(value)

class Void:
    """Void (void)"""
    def __init__(self):
        self.value = None  # always None

    def __repr__(self):
        return "void"

    def to_ctypes(self):
        """Returns CTypes compatible instance."""
        import ctypes
        return None

class Pointer:
    """A pointer (*)"""
    def __init__(self, base_type, value=None):
        self.base_type = base_type
        self.value = value  # can be None initially

    def deref(self):
        """Dereferences the pointer (&ptr)"""
        if self.value is None:
            raise ValueError("Pointer is NULL")
        return self.value

    def set(self, value):
        """Sets a value"""
        if not isinstance(value, self.base_type):
            raise TypeError(f"Expected {self.base_type}, got {type(value)}")
        self.value = value

    def __repr__(self):
        addr = hex(id(self.value)) if self.value is not None else "NULL"
        return f"<Pointer to {self.base_type.__name__} at {addr}>"

    def to_ctypes(self):
        """Returns CTypes compatible instance."""
        import ctypes
        if self.value is None:
            return ctypes.c_void_p(None)
        elif hasattr(self.value, "to_ctypes"):
            return ctypes.pointer(self.value.to_ctypes())
        else:
            return ctypes.c_void_p(id(self.value))  # fallback

class VoidPointer(Pointer):
    """A Void Pointer (void*)"""
    def __init__(self, value=None):
        super().__init__(Void, value)

class CPrebuilt:
    def __init__(self):
        self.functions = {}
        self.macros = {}
        self.cond_stack = []
        self.typedefs = {}
        self.enums = {}
        self.structs = {}
        self.inc_dirs = []

    def include(self, path:str, include_dirs: Union[str, list]="", ignore_includes:bool=False):
        """Includes a header file and auto resolves all dependencies, functions, macros and conditionals
        
        Parameters:
            path (str): Path to the file to be resolved. Path can be absolute or relative to cwd or relative to any of the paths in include_dirs (NOTE: INCLUDE_DIRS ARE SAVED PER INSTANCE).
            include_dirs (str | list): Either a single path to any include directory or a list containing multiple include directory paths. Defaults to ''
            ignore_includes (bool): Wether to ignore include statements. Defaults to False."""
        if isinstance(include_dirs, str):
            self.inc_dirs.append(include_dirs)
        elif isinstance(include_dirs, list):
            self.inc_dirs = include_dirs
        
        self.inc_dirs.append("")
        exists = False
        
        for dir_ in self.inc_dirs:
            path = os.path.join(dir_, path)
            if os.path.exists(path):
                exists = True
                break
            
        if not exists:
            raise FileNotFoundError(f"Header file {path} was not found!")
        # Parse header and store metadata for auto-resolve
        content = ""
        with open(path, "r") as f:
            f.seek(0)
            content = f.read()

        self.parse(content, ignore_includes)

    def parse(self, text:str, ignore_includes:bool=False):
        """Parse C header text for functions, macros, and conditionals
        
        Parameters:
            text (str): The data to parse.
            ignore_includes (bool): Wether to ignore include statements. Defaults to False."""
        lines = text.splitlines()
        buffer = []
        in_comment = False
        for line in lines:
            line = line.strip()
            
            if line.endswith("*/") and in_comment:
                in_comment = False
                continue
            
            if in_comment:
                continue
        
            if not line or line.startswith("//"):
                continue  # skip empty lines and single line comments
            
            if line.startswith("/*") and not in_comment:
                in_comment = True
                continue

            # Preprocessor
            if line.startswith("#define"):
                parts = line.split(maxsplit=2)
                key = parts[1]
                val = parts[2] if len(parts) > 2 else "1"
                self.macros[key] = val
                continue
            elif line.startswith("#undef"):
                key = line.split()[1]
                self.macros.pop(key, None)
                continue
            elif line.startswith("#ifdef"):
                self.cond_stack.append(line.split()[1])
                continue
            elif line.startswith("#ifndef"):
                self.cond_stack.append("!" + line.split()[1])
                continue
            elif line.startswith("#else"):
                if self.cond_stack: 
                    self.cond_stack[-1] = "!" + self.cond_stack[-1]
                continue
            elif line.startswith("#endif"):
                if self.cond_stack: 
                    self.cond_stack.pop()
                continue
            elif line.startswith("#include") and not ignore_includes:
                parts = line.split()
                if len(parts) < 2:
                    continue
                file = parts[1].replace('"', "").replace("<", "").replace(">", "")
                self.include(file)

            # Function detection
            buffer.append(line)
            if line.endswith(";") and not line.startswith("typedef"):
                err = self._parse_function(" ".join(buffer))
                if not err == "ERROR":
                    buffer = []
                
            # Typedef/enums/structs
            b1 = buffer[0] if len(buffer) > 0 else ""
            bend = buffer[-1] if len(buffer) > 0 else ""
            if (bend.endswith("}") or bend.endswith(";")) and (b1.startswith("typedef") or (b1.startswith("enum") or b1.startswith("struct"))):
                if b1.startswith("struct"):
                    buffer = []
                    continue
                err = self._parse_tes(" ".join(buffer))
                if not err == "ERROR":
                    buffer = []
            
    def _join_function_lines(self, text:str):
        buf, out = [], []
        for line in text.splitlines():
            buf.append(line.strip())
            if ";" in line:         # function decl ends
                out.append(" ".join(buf))
                buf = []
        return out

    def _parse_function(self, decl:str):
        decl = decl.rstrip(';').strip()
        if '(' not in decl or ')' not in decl:
            return "ERROR"

        # Split into return type, name, args
        ret_and_name, args = decl.split('(', 1)
        
        if not args.startswith(");") and (not args.startswith("void);") and (not args.startswith(")") and not args.startswith("void)"))):
            args = args.rsplit(')', 1)[0]
        else:
            args = "void"
        ret_parts = ret_and_name.strip().split()
        name = ret_parts[-1].lstrip('*')  # handle pointers
        ret_type = " ".join(ret_parts[:-1])
        extern = False
        
        if ret_type.startswith("extern "):
            extern = True
            ret_type = ret_type.replace("extern ", "")
        
        args_list = [a.strip() for a in args.split(',') if a.strip() and a.strip() != "void"]
        
        args_dict = {}
        for item in args_list:
            if not item or item == "void":
                continue
            arg = item.split(" ")
            if len(arg) < 2:
                continue # Skip broken/voids
            args_dict[arg[1]] = arg[0]

        self.functions[name] = {
            "return": ret_type,
            "name": name,
            "args": args_dict,
            "extern": extern,
        }
    
    def _parse_tes(self, text:str) -> None:
        if not text.startswith("typedef") and (not text.startswith("struct") and not text.startswith("enum")):
            return "ERROR"
        
        if text.startswith("typedef"):
            return self._parse_typedef(text)
        if text.startswith("struct"):
            return self._parse_struct(text)
        if text.startswith("enum"):
            return self._parse_enum(text)
            
    def _parse_typedef(self, text:str):
        if not text.startswith("typedef"):
            return "ERROR"
        
        text_arr = text.replace("\n", "").split(" ")
        if len(text_arr) < 3:
            return "ERROR"
        type_ = text_arr[1]
        name = ""
        val = None
        if len(text_arr) > 3:
            if type_ not in ["enum", "struct"]:
                return "ERROR"
            
            if type_ == "enum":
                enum = "".join(text_arr).replace("typedef", "").replace("enum", "")
                if not enum.startswith("{"):
                    return "ERROR"
                
                enum_vals = enum.replace("{", "").split(",")
                if "}" in enum_vals[-1]:
                    final, name = enum_vals[-1].split("}")
                    name = name.replace(";", "")
                    enum_vals[-1] = final
                
                val = enum_vals
        else:
            name = text_arr[2]
            name = name.replace(";", "")
            
        self.typedefs[name] = {
            "name": name,
            "value": val,
            "type": type_
        }

    def get_function(self, name:str) -> dict:
        """Gets a function from the resolved functions"""
        return self.functions.get(name, None)

    @property
    def resolved_macros(self):
        return self.macros

    @property
    def conditional_stack(self):
        return self.cond_stack

    @property
    def resolved_functions(self):
        return self.functions
        
    @property
    def resolved_typedefs(self):
        return self.typedefs
    
class LibraryWrapper:
    def __init__(self, path):
        if not os.path.exists(path):
            raise FileNotFoundError(f"Library {path} not found")
        self.path = path
        self._lib = ctypes.CDLL(path)  # ctypes handles platform ABI
        self._func_cache = {} # caches resolved functions

    def __getattr__(self, name):
        # auto-resolve happens here
        if name in self._func_cache:
            return self._func_cache[name]

        try:
            func = getattr(self._lib, name)
        except AttributeError:
            raise AttributeError(f"Function {name} not found in {self.path}")

        # Auto-resolve signature
        func.argtypes, func.restype = self._auto_resolve(name)

        def wrapper(self:LibraryWrapper, *args):
            conv_args = [self._to_ctypes(arg) for arg in args]
            res = func(*conv_args)
            return self._from_ctypes(res, func.restype)

        self._func_cache[name] = func
        return func

    def _auto_resolve(self, func_name):
        return ([], ctypes.c_int)

    def _to_ctypes(self, arg):
        if hasattr(arg, "to_ctypes"):
            return arg.to_ctypes()
        else:
            return arg

    def _from_ctypes(self, value, restype):
        return value


"""Graphical User Interface Backends and SubModules"""

import platform

import sys
import os

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions import HyperOut
from phardwareitk import Extensions

from typing import *

def get_platform() -> str:
    return platform.system().lower()

#Check Platform
OS:str = get_platform()
Linux:bool = False
Windows:bool = False
Darwin:bool = False
Unix:bool = False
Unknown_os:bool = True

QUIT = 256 #SDL_QUIT

if OS == "windows":
    Windows = True
    Unknown_os = False
elif OS == "linux":
    Linux = True
    Unknown_os = False
elif OS == "darwin":
    Darwin = True
    Unknown_os = False
elif OS == "unix":
    Unix = True
    Unknown_os = False
else:
    Unknown_os = True

# Constants
WM_DESTROY:int = 2
WM_PAINT:int = 15


"""Pyrion GPU"""
from typing import *

from .constants import *

_gpu_context = None

def base_pyrion_msg_handler(stream: PSTREAM, *args) -> Any:
    """Basic Pyrion Message Handler"""
    stream.value(*args)

def base_pyrion_err_stream(*args) -> None:
    """Default Pyrion Error stream"""
    from phardwareitk.Extensions.HyperOut import printH
    from phardwareitk.Extensions import Color, TextFont

    printH(*args, FontEnabled=True, Font=TextFont(font_color=Color("red"), Bold=True))

class PyrionGPU:
    """
    Represents the Virtual GPU. Handles memory, shaders, execution, and more...
    
    NOTE: Uses the CPU, for GPU usage, please resort to using PyrionGPU driver for AOS++, CPU is the only option because Operating Systems like Window do not give GPU access without going through layers like OpenGL/Vulkan/DirectX/etc which makes it more complex and can't be highly feature rich because it will in end have to rely on already made systems.
    """
    def __init__(self):
        self.initialized = False

        self.nextalloc = {
            "array": 0
        }

        self.memory = bytearray()
        self.framebuffer = bytearray()
        self.meminfo = {}
        self.shaders = {}
        self.kernels = {}
        self.context_info = self.detect_hardware()

        self.outstream = PSTREAM(object, print)
        self.instream = PSTREAM(object, input)
        self.errstream = PSTREAM(object, base_pyrion_err_stream)

        self.iohandler = base_pyrion_msg_handler

    def detect_hardware(self) -> Dict[str, Any]:
        """Detect the GPU capabilities and resources available"""
        return {
            "name": "Pyrion Virtual GPU",
            "vendor": "Pheonix Studios",
            "max_memory": 1024 * 1024 * 1024, # 1 GB
            "supported_shaders": ["vertex", "fragment", "compute"]
        }
    
    def print(self, *args) -> None:
        """Outputs the *args to the output stream"""
        self.iohandler(self.outstream, *args)

    def fprint(self, stream: PSTREAM, *args) -> None:
        """Outputs at specified stream"""
        stream.value(*args)

    def change_stream(self, stream: PSTREAM, stream_type: PSTREAMID) -> None:
        """Change the current <stream_type> to <stream>"""
        if stream_type == PSTREAMOUT:
            self.outstream = stream
        elif stream_type == PSTREAMIN:
            self.instream = stream
    
    def initialize(self, memory_size: PLONG) -> None:
        """Initialize the GPU with specified memory size"""
        if self.initialized: self.fprint(self.errstream, "Pyrion is already initialized!")
        self.print("Initializing Pyrion VGPU...")
        self.initialized = True
        self.memory = bytearray(int(memory_size))
        self.print("Pyrion VGPU is initialized!")

    def allocate_mem(self, value: PLONG, mem_type:PMEMID) -> PMEMORY:
        """Allocates Pyrion memory for a object, returns PMEMORY (*long)"""
        if mem_type == PMEM_ARRAY:
            ret = PMEMORY(id(self.memory) + self.nextalloc["array"])
            self.meminfo[ret.address] = {
                "type": int(mem_type),
                "size": int(value)
            }

            self.nextalloc["array"] += value
            self.print(f"Allocated memory of size {hex(value)}")
            return ret

    def free_mem(self, memory: PMEMORY) -> None:
        mem = self.meminfo.get(memory.address, None)
        if not mem:
            self.fprint(self.errstream, f"No such allocated memory at address {hex(memory.address)}")
            return None
        
        if mem["type"] == int(PMEM_ARRAY):
            self.write_mem(memory, pyrion_ptr(bytes, b"\x00"*mem["size"]), POFFSET(0))
        
    def write_mem(self, ptr: PMEMORY, data_ptr: PBYTES, offset: POFFSET=POFFSET(0)):
        """Write data into VRAM at pointer offset"""

        data = data_ptr.value

        start = ptr.address - id(self.memory) + offset
        end = start + len(data)
        self.memory[start:end] = data

    def read_mem(self, ptr: PMEMORY, size: PLONG, offset: POFFSET=POFFSET(0)) -> bytes:
        start = ptr.address - id(self.memory) + offset
        return self.memory[start:start+size]
    
    def create_ctx(self, ctx_info: PCTX_INFO) -> PCONTEXT:
        """Creates a context based on the context info, this context have to then be attached though"""
        ctx = PCONTEXT()
        ctx.handle = PHANDLE(0) # None
        ctx.pixel_size = ctx_info.pixel_size
        ctx.pixel_depth = ctx_info.pixel_depth
        ctx.fb_initialclr = ctx_info.initcolor
        ctx.fb_width = ctx_info.width
        ctx.fb_height = ctx_info.height

        return ctx
    
    


from typing import *
import os
import sys
import time

if not os.path.dirname(os.path.abspath(__file__)) in sys.path:
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from C_Old import *

do_print_exception = False

def set_exception_print(val:bool) -> None:
    """True for printing exception, False for otherwise"""
    global do_print_exception
    do_print_exception = val
    
def print_exception(exception) -> None:
    """Prints exception provided based on the do_print_exception, whose value is setted by set_exception_print function"""
    global do_print_exception
    if do_print_exception:
        print(exception)

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

_IO_FILE = {
    "_IO_read_ptr": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_read_end": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_read_base": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_write_base": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_write_ptr": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_write_end": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_buf_base": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_IO_buf_end": {
        "type": Pointer,
        "ptr_to": Char,
        "value": None
    },
    "_fileno": {
        "type": Int,
        "value": None
    },
    "_blksize": {
        "type": Int,
        "value": None
    }
}
"""struct _IO_FILE {
   char  *_IO_read_ptr;   /* Current read pointer */
   char  *_IO_read_end;   /* End of get area. */
   char  *_IO_read_base;  /* Start of putback and get area. */
   char  *_IO_write_base; /* Start of put area. */
   char  *_IO_write_ptr;  /* Current put pointer. */
   char  *_IO_write_end;  /* End of put area. */
   char  *_IO_buf_base;   /* Start of reserve area. */
   char  *_IO_buf_end;   /* End of reserve area. */
   int   _fileno;
   int   _blksize;
};

typedef struct _IO_FILE FILE;"""

FILE = Struct[_IO_FILE]

def fopen(path:Union[Pointer[Char], str], mode:str) -> Union[Pointer[FILE], int]:
    """fopen - Open a file and return a FILE struct pointer

    Arguments:
        path (char*):
            Path to the file
        mode (char*):
            Mode to open the file with (e.g. 'r', 'w', 'rb', ...)

    Returns: Pointer to FILE (_IO_FILE) struct or -1 on failure, -2 on file not found, -3 on permission error"""
    filename = get_string(path)
    mode = get_string(mode)

    fd = None

    try:
        fd = open(filename, mode)
    except PermissionError as pe:
        print_exception(pe)
        return -3

    if fd is None:
        return -1

    buffer_size = 4096 # Default
    statvfs = None
    block_size = None

    try:
        statvfs = os.statvfs(filename)
        block_size = statvfs.f_bsize
    except:
        block_size = buffer_size

    # allocate memory for buffer
    buf_base = malloc(buffer_size)

    # Allocate FILE struct
    file = Struct(_IO_FILE)
    # set read values first
    file.set("_IO_read_ptr", buf_base)
    file.set("_IO_read_end", buf_base) # empty buffer at the start
    file.set("_IO_read_base", buf_base)

    # set write values
    file.set("_IO_write_base", buf_base)
    file.set("_IO_write_ptr", buf_base)
    file.set("_IO_write_end", buf_base.address + buffer_size)

    # set buffer values
    file.set("_IO_buf_base", buf_base)
    file.set("_IO_buf_end", buf_base.address + buffer_size)

    # set file descriptor and block size
    try:
        file.set("_fileno", fd.fileno())
    except Exception as e:
        print_exception(e)
        file.set("_fileno", -1)

    file.set("_blksize", block_size)

    # Store the file descriptor in a global dictionary to keep track of open files
    if "_open_files" not in globals():
        global _open_files
        _open_files = {}
    _open_files[file.access("_fileno")] = fd

    return Pointer(Struct, file, True) # Return

def fclose(file_:Pointer[FILE]) -> int:
    """fclose - Close a file and free its memory

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: 0 on success, -1 on failure"""
    try:
        # Dereference the pointer
        file: Struct = file_.dereference()
        fileno = file.access("_fileno")

        # Get the file descriptor from the global dictionary
        fd:TextIO = _open_files.get(fileno, None)

        if fd is None:
            return -1

        # Close the file
        fd.close()

        # Free the buffer memory
        free(file_) # Free the FILE struct memory
        del file
        del file_

        # Remove the file descriptor from the global dictionary
        del _open_files[fileno]

        return 0
    except Exception as e:
        print_exception(e)
        return -1

def ftell(file: Pointer[FILE]) -> int:
    """ftell - Get the current position in the file

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: Current position in the file or -1 on failure"""
    try:
        # Dereference the pointer
        file_struct: Struct = file.dereference()
        fd = _open_files.get(file_struct.access("_fileno"))
        if not fd:
            del file_struct
            return -1
        out = fd.tell()
        del file_struct
        return out
    except Exception as e:
        print_exception(e)
        return -1

def fflush(file: Pointer[FILE]) -> int:
    """fflush - Flush the output buffer of a file

    Arguments:
        file (FILE*):
            Pointer to the FILE struct

    Returns: 0 on success, -1 on failure"""
    try:
        # Dereference the pointer
        file_struct: Struct = file.dereference()
        fd:TextIO = _open_files.get(file_struct.access("_fileno"), None)

        if fd is None:
            return -1

        # Flush the file
        fd.flush()
        del file_struct
        return 0
    except Exception as e:
        print_exception(e)
        return -1

def fseek(file: Pointer[FILE], offset: int, whence: int) -> int:
    """fseek - Set the file position indicator for the stream

    Arguments:
        file (FILE*):
            Pointer to the FILE struct
        offset (int):
            Offset to set the position to
        whence (int):
            Position from which to set the offset (e.g. SEEK_SET, SEEK_CUR, SEEK_END)

    Returns: 0 on success, -1 on failure"""
    try:
        # Dereference the pointer
        file_struct: Struct = file.dereference()
        fd:TextIO = _open_files.get(file_struct.access("_fileno"), None)

        if fd is None:
            return -1

        fd.seek(offset, whence)
        del file_struct
        return 0
    except Exception as e:
        print_exception(e)
        return -1

def fread(dest: Pointer[Void], size: Union[int, Size_t], nmemb: Union[int, Size_t], file_ptr: Pointer[FILE]) -> int:
    """fread - Read data from a file

    Args:
        dest (Pointer[Void]): The destination pointer where data will be stored
        size (int): size of each element to read
        nmemb (int): number of elements to read
        file_ptr (Pointer[FILE]): Pointer to the FILE struct

    Returns:
        int: Number of elements successfully read, or -1 on failure
    """
    if isinstance(size, Size_t):
        size = size.bytes

    if isinstance(nmemb, Size_t):
        nmemb = nmemb.bytes

    if not isinstance(size, int) or not isinstance(nmemb, int):
        raise TypeError("size and nmemb must be int or Size_t")

    try:
        # Dereference the pointer
        file:FILE = file_ptr.dereference()
        total_bytes = size * nmemb

        fileno = file.access("_fileno")
        fd:TextIO = _open_files.get(fileno, None)
        if fd is None:
            return -1

        data = fd.read(total_bytes)
        if not data:
            return 0

        # Write data to the destination pointer
        write(dest, data, total_bytes, no_meta=True)

        # Update the read pointer
        buf_base = file.access("_IO_buf_base")
        file.set("_IO_read_ptr", Pointer(buf_base.address + len(data)))
        file.set("_IO_read_end", Pointer(buf_base.address + total_bytes))

        del file

        return len(data) // size

    except Exception as e:
        print_exception(e)
        return -1

def fwrite(src: Union[Pointer[Void], str, bytes], size: Union[int, Size_t], nmemb: Union[int, Size_t], file_ptr: Pointer[FILE], chunk_size:int=4096) -> int:
    """fwrite - Write data to a file
    
    Args:
        src (Pointer[Void]): The source pointer containing data to write
        size (int): size of each element to write
        nmemb (int): number of elements to write
        file_ptr (Pointer[FILE]): Pointer to the FILE struct
        chunk_size (int): Writing files in chunk to prevent overuse of memory. Defaults to 4096 bytes / 4KB

    Returns:
        int: Number of elements successfully written, or -1 on failure
    """
    if isinstance(size, Size_t):
        size = size.bytes

    if isinstance(nmemb, Size_t):
        nmemb = nmemb.bytes

    if not isinstance(size, int) or not isinstance(nmemb, int):
        raise TypeError("size and nmemb must be int or Size_t")

    try:
        # Dereference the pointer
        file:FILE = file_ptr.dereference()
        total_bytes = size * nmemb

        fileno = file.access("_fileno")
        fd:TextIO = _open_files.get(fileno, None)
        if fd is None:
            return -1
        
        written = 0
        srcaddr = None
        if isinstance(src, Pointer):
            srcaddr = src.pointer_address
        
        while written < total_bytes:
            # Read data from the source pointer
            to_write = min(chunk_size, total_bytes - written)
            
            if isinstance(src, Pointer):
                data = get_mem(srcaddr + written, to_write)
            else:
                data = src
    
            if not 'b' in fd.mode:
                try:
                    data = data.decode('utf-8')
                except UnicodeDecodeError:
                    return -1
    
            # Write data to the file
            fd.write(data)
            written += to_write

        # Update the write pointer
        buf_base = file.access("_IO_buf_base")
        file.set("_IO_write_ptr", Pointer(buf_base.address + written))

        del file

        return len(data) // size

    except Exception as e:
        print_exception(e)
        return -1

def remove(filename: Union[Pointer[Char], str]) -> int:
    """remove - Remove a file

    Arguments:
        filename (char*):
            Path to the file to remove

    Returns: 0 on success, -1 on failure"""
    try:
        os.remove(get_string(filename))
        return 0
    except FileNotFoundError as fe:
        print_exception(fe)
        return -2
    except PermissionError as pe:
        print_exception(pe)
        return -3
    except Exception as e:
        print_exception(e)
        return -1

def rename(old_filename: Union[Pointer[Char], str], new_filename: Union[Pointer[Char], str]) -> int:
    """rename - Rename a file

    Arguments:
        old_filename (char*):
            Path to the file to rename
        new_filename (char*):
            New path for the file

    Returns: 0 on success, -1 on failure"""
    try:
        os.rename(get_string(old_filename), get_string(new_filename))
        return 0
    except FileNotFoundError as fe:
        print_exception(fe)
        return -2
    except PermissionError as pe:
        print_exception(pe)
        return -3
    except Exception as e:
        print_exception(e)
        return -1




import sys

from .types import *
from .types import _PyrionPointer

if sys.maxsize > 2**32:
    PBITS = 64
else:
    PBITS = 32

# Typedefs
PBYTE = pyrion_int8
PSHORT = pyrion_int16
PINT = pyrion_int32
_PLONG = pyrion_int64 if PBITS == 64 else pyrion_int32 # Did this to fix typing issue
PLONG = _PLONG
PUBYTE = pyrion_uint8
PUSHORT = pyrion_uint16
PUINT = pyrion_uint32
_PULONG = pyrion_uint64 if PBITS == 64 else pyrion_uint32 # same reason
PULONG = _PULONG

PBYTES = pyrion_ptr[bytes]

PNULL = pyrion_void

PSTREAMID = PUINT
PYID = PUINT
PSHADERID = PUINT
PMEMID = PLONG

PSIZE = PUINT
PLEN = PUINT
POFFSET = PLONG

PVOID = pyrion_ptr[PNULL]
PSTREAM = pyrion_ptr[object]
PMEMORY = _PyrionPointer[PMEMID]
PBUFFER = pyrion_ptr[PULONG]

PSTREAMOUT = PSTREAMID(1)
PSTREAMIN = PSTREAMID(0)

PMEM_ARRAY = PMEMID(0)
PMEM_BUFF = PMEMID(1)
PMEM_VERTEX = PMEMID(2)

PIXEL_SIZE = PUINT
PIXEL_RGB = PIXEL_SIZE(3)
PIXEL_RGBA = PIXEL_SIZE(4)

PHANDLE = PLONG

class PCONTEXT(pyrion_structure):
    _struct_ = [
        ("handle", PHANDLE),
        ("pixel_size", PIXEL_SIZE),
        ("pixel_depth", PUBYTE),
        ("fb_initialclr", PUINT),
        ("fb_width", PUINT),
        ("fb_height", PUINT)
    ]

class PCTX_INFO(pyrion_structure):
    _struct_ = [
        ("ctx", PCONTEXT),
        ("pixel_size", PIXEL_SIZE),
        ("pixel_depth", PUBYTE),
        ("width", PUINT),
        ("height", PUINT),
        ("initcolor", PUINT),
    ]